#
# This module reads files in Axon Binary File format v1 or v2 used by Axon
# Technologies and Molecular Devices.
# The v1 format was used until Clampex version 9.
# Clampex 10 and onwards use the v2 format.
#
# WARNING: This file hasn't been extensively tested.
#
# About ABF
# ---------
# pClamp version 10 introduced a new .abf file format ABF2, with format version
#  numbers 2.0 and up. Older version are referred to a version 1 (even though
#  the actual version number may be, for example, 1.6).
# The version 2 format uses a variable sized header making it a little trickier
#  to read but appears to be more structured overal than the first version.
#
# Trials, runs, sweeps and channels
# ---------------------------------
# In pClamp version 10 and higher, each recorded data segment is termed a
#  'sweep'. In older versions the same concept is called an 'episode'.
# The data in a sweep contains the recordings from one or more channels. The
#  number of channels used throughout a single file is constant. In other
#  words, channel 1 in sweep 1 is a recording from the same channel as channel
#  1 in sweep 10.
# A set of sweeps is called a run. Both abf versions 1 and 2 contain a variable
#  indicating the number of sweeps per run. In abf1 this is found in the
#  header, in abf2 it is stored in a separate 'protocol' section.
# It is possible to record multiple runs, each containing an equal number of
#  sweeps, each containing an equal number of channels. However, in this case
#  the data from run to run is averaged and only a single averaged run is
#  saved. This means there is never more than 1 run in a file, even though this
#  data may have been obtained during multiple runs.
# A set of runs is termed a 'trial'. Each file contains a single trial.
#
# Acquisition modes
# -----------------
# pClamp uses five acquisition modes:
#  Gap-free mode
#    Data is recored continuously, without any interruptions
#  Variable-length events mode
#    Data is recorded in bursts of variable length (for example whenever some
#    condition is met)
#  Fixed-length events mode
#    Data is recored in bursts of fixed length, starting whenever some
#    condition is met. Multiple bursts (sweeps, or episodes in pClamp <10
#    terminology) may overlap.
#  High-speed oscilloscope mode
#    Like fixed-length events mode, but sweeps will never overlap
#  Episodic stimulation mode
#    Some stimulation is applied during which the resulting reaction is
#    recorded. The resulting dataset consists of non-overlapping sweeps.
#
# Stimulus waveforms
# ------------------
# A stimulus signal in pClamp is termed a 'waveform'. Each waveform is divided
#  into a series of steps, ramps or pulse trains. Such a subsection is called
#  an 'epoch'. The protocol section of a file defines one or more stimuli, each
#  containing a list of epochs.
#
#---------------------------------  license  ----------------------------------
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
#---------------------------------  credits  ----------------------------------
#
# A lot of this code uses the (sadly somewhat outdated) information made public
# by Axon, e.g. at https://support.moleculardevices.com/s/article/
#                               Axon-pCLAMP-ABF-File-Support-Pack-Download-Page
# This information comes without a specific license, but states that
# "Permission is granted to freely use, modify and copy the code in this file."
#
# In addition, this module was in part derived from an early version of the
# Neo package for representing electrophysiology data, specifically from a
# Python module authored by sgarcia and jnowacki.
# Neo can be found at: http://neuralensemble.org/trac/neo
#
# The Neo package used was licensed using the following BSD License:
#
#----------------------------------  start  -----------------------------------
# Copyright (c) 2010-2012, Neo authors and contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# Neither the names of the copyright holders nor the names of the contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#-----------------------------------  end  ------------------------------------
# The code used in Neo is itself derived from the publicly contributed matlab
#  script abf2load, again licensed under BSD. The original notice follows
#  below:
#----------------------------------  start  -----------------------------------
# Copyright (c) 2009, Forrest Collman
# Copyright (c) 2004, Harald Hentschke
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#-----------------------------------  end  ------------------------------------
# The abf2load script is available from:
#  http://www.mathworks.com/matlabcentral/fileexchange/22114-abf2load
#------------------------------------------------------------------------------
# Information - but no direct code - from the matlab script get_abf_header.m
# was also used: http://neurodata.hg.sourceforge.net/hgweb/neurodata/neurodata/
#------------------------------------------------------------------------------
import datetime
import os
import struct
import warnings

import numpy as np

import myokit
import myokit.formats


# Encoding for text parts of files
_ENC = 'latin-1'


class AbfFile(myokit.formats.SweepSource):
    """
    Represents a read-only Axon Binary Format file (``.abf``), stored at the
    location pointed to by ``filepath``.

    Files in the "ABF" format and the newer "ABF2" format can be read. If the
    given ``filepath`` ends in ``.pro`` a protocol file is assumed. This
    assumption can be overruled by setting the ``is_protocol_file`` argument
    to either ``True`` or ``False``.

    Data in ABF files is recorded in *sweeps*, where each sweep contains one or
    more *channels* with recorded (A/D) data. In addition, zero or more output
    waveforms may be defined (also called "protocol" or D/A channels). Where
    possible, the :class`AbfFile` class will convert these embedded protocols
    to time series and include them as additional channels.

    For example::

        abf = AbfFile('some_file.abf')
        for sweep in abf:
            for channel in sweep:
                print(channel.name())
            break

    might show

        IN 0
        10xVm
        Cmd 0

    where the first two channels are recorded A/D channels and the final one is
    a reconstructed D/A output channel.

    Sweeps and channels are represented by :class:`Sweep` and :class:`Channel`
    objects respectively, and these can be used to obtain the data from a
    file::

        abf = AbfFile('some_file.abf')
        for sweep in abf:
            for channel in sweep:
                plot(channel.times(), channel.values())

    In addition the ``AbfFile`` class implements the
    :class`myokit.formats.SweepSource` interface. Note that this interface
    treats A/D and D/A as separate things, so :meth:`channel_count` returns the
    number of A/D channels, not the total number of channels in a
    :class:`Sweep` object (which can include D/A channels).

    Support notes:

    - Protocol (D/A) conversion is only supported for "episodic stimulation"
      with constant valued steps (so e.g. no ramps) and without "user lists".
    - Protocols with more than one sampling rate are not supported.
    - The publicly available information on the ABF format is not great, so
      there will be several other issues and shortcomings.

    Arguments:

    ``filepath``
        The path to load the data from. Data will be read into memory
        immediately upon construction.
    ``is_protocol_file``
        If set to ``True``, no attempt to read A/D data will be made and only
        D/A "protocol" information will be read. If left at its default value
        of ``None`` files with the extension ``.pro`` will be recognized as
        protocol files.

    """
    def __init__(self, filepath, is_protocol_file=None):
        # The path to the file and its basename
        filepath = str(filepath)
        self._filepath = os.path.abspath(filepath)
        self._filename = os.path.basename(filepath)

        # Read as protocol file yes?
        if is_protocol_file is None:
            self._is_protocol_file = os.path.splitext(filepath)[1] == '.pro'
        else:
            self._is_protocol_file = bool(is_protocol_file)

        # Cached string-to-unit conversions
        self._unit_cache = {}

        # Read the file header as an ordered dictionary
        self._header = None
        self._strings = None
        self._version = None
        self._version_str = None
        self._read_1_header()

        # Read the time of recording
        self._datetime = None
        self._read_2_time_of_recording()

        # Read the protocol information
        self._n_adc = None
        self._n_dac = None

        self._rate = None
        self._mode = None

        self._number_of_trials = None
        self._trial_start_to_start = None
        self._runs_per_trial = None
        self._run_start_to_start = None
        self._sweeps_per_run = None
        self._sweep_start_to_start = None
        self._samples_per_channel = None

        # To be able to treat v1 and v2 slightly more easily, we define 3
        # functions read epoch info from episodic stimulation protocols
        self._epoch_functions = None

        # Not all D/A channels can be converted, so we maintain an array with
        # the original indices of the channels in the da_sweeps. (Note that
        # this differs again from the "index" labels for user display, which
        # are stored in the channels themselves).
        self._dac_indices = []

        # Read protocol information and create empty sweep objects
        # Sweeps contain both A/D channels and D/A reconstructions. Some files
        # will have A/D but no (or no supported) D/A. Conversely protocol files
        # will have D/A only. So all in one sweep is easiest.
        self._sweeps = None

        self._read_3_protocol_information()

        # Read and calculate conversion factors for integer data in ADC
        self._adc_factors = None
        self._adc_offsets = None
        self._read_4_ad_conversion_factors()

        # Read the A/D channel data and add it to the sweeps
        if self._n_adc:
            self._read_5_ad_data()

        # Reconstruct D/A signals as additional channels and add to the sweeps
        self._read_6_da_reconstructions()

        # Copy channel names and units, for easier SweepSource implementation
        self._ad_names = {}
        self._da_names = {}
        self._ad_units = []
        self._da_units = []
        for sweep in self._sweeps:
            for i, channel in enumerate(sweep[:self._n_adc]):
                self._ad_names[channel.name()] = i
                self._ad_units.append(channel.unit())
            for i, channel in enumerate(sweep[self._n_adc:]):
                self._da_names[channel.name()] = i
                self._da_units.append(channel.unit())
            break

    def _read_1_header(self):
        """ Read the file header. """

        def read_f(f, form, offset=None):
            """ Read and unpack a file section in the format ``form``. """
            form = str(form)
            if offset is not None:
                f.seek(offset)
            return struct.unpack(form, f.read(struct.calcsize(form)))

        def ups(val):
            """
            Parses tuples, unpacking single values:

            1. Tuples with a single entry are converted to that entry (e.g.
               ``(5, )`` gets turned into ``5``.
            2. Any bytes objects are converted to string.
            3. Strings containing only \\x00 are replaced by ``None``

            """
            # Convert bytes
            values = [0] * len(val)
            for i, v in enumerate(val):
                if isinstance(v, bytes):
                    v = v.decode(_ENC).strip()
                    # Handle long \x00 lists
                    if v and ord(v[0]) == 0:
                        return None
                values[i] = v
            del val

            # Unpack single
            if len(values) == 1:
                return values[0]
            return values

        with open(self._filepath, 'rb') as f:

            # Get ABF Format version (pClamp < 10 is version 1, after is
            # version 2)
            sig = f.read(4).decode(_ENC)
            if sig == 'ABF ':
                version = 1
            elif sig == 'ABF2':
                version = 2
            else:
                raise NotImplementedError('Unknown ABF Format "{sig}".')

            # Gather header fields
            header = {}
            for key, offset, form in HEADER_FIELDS[version]:
                header[key] = ups(read_f(f, form, offset))

            # Get uniform file version number
            if version < 2:
                self._version = np.round(header['fFileVersionNumber'], 5)
                self._version_str = str(self._version)
            else:
                v = header['fFileVersionNumber']
                self._version = v[3]
                self._version_str = '.'.join([str(v) for v in reversed(v)])

            # Get file start time in seconds
            if version < 2:
                header['lFileStartTime'] += (
                    header['nFileStartMillisecs'] / 1000)
            else:
                header['lFileStartTime'] = header['uFileStartTimeMS'] / 1000

            if version < 2:

                # Version 1: Only read tags
                tags = []
                for i in range(header['lNumTagEntries']):  # pragma: no cover
                    # Cover pragma: Don't have appropriate test file
                    f.seek(header['lTagSectionPtr'] + i * 64)
                    tag = {}
                    for key, form in ABF2_TAG_INFO_DESCRIPTION:
                        tag[key] = ups(read_f(f, form))
                    tags.append(tag)
                header['tags'] = tags
                self._strings = []

            else:

                # Version 2
                # Find location of file sections
                sections = {}
                for i, s in enumerate(ABF2_FILE_SECTIONS):
                    index, data, length = read_f(f, 'IIl', 76 + i * 16)
                    sections[s] = {}
                    sections[s]['index'] = index
                    sections[s]['data'] = data
                    sections[s]['length'] = length
                header['sections'] = sections

                # String section contains channel names and units
                f.seek(sections['Strings']['index'] * BLOCKSIZE)
                strings = f.read(sections['Strings']['data'])

                # Starts with header we need to skip
                # DWORD dwSignature;    4 bytes
                # DWORD dwVersion;      4 bytes
                # UINT  uNumStrings;    4 bytes
                # UINT  uMaxSize;       4 bytes
                # ABFLONG  lTotalBytes; 4 bytes
                # UINT  uUnused[6];     24 bytes
                # Total: 44 bytes
                strings = strings[44:]

                # C-style string termination
                strings = strings.split(b'\x00')
                strings = [s.decode(_ENC).strip() for s in strings]
                self._strings = strings

                # Read tag section
                tags = []
                offs = sections['Tag']['index'] * BLOCKSIZE
                size = sections['Tag']['data']
                for i in range(sections['Tag']['length']):  # pragma: no cover
                    # Cover pragma: Don't have appropriate test file
                    f.seek(offs + i * size)
                    tag = {}
                    for key, form in ABF2_TAG_INFO_DESCRIPTION:
                        tag[key] = ups(read_f(f, form))
                    tags.append(tag)
                header['tags'] = tags

                # Read protocol section
                protocol = {}
                offs = sections['Protocol']['index'] * BLOCKSIZE
                f.seek(offs)
                for key, form in ABF2_PROTOCOL_FIELDS:
                    protocol[key] = ups(read_f(f, form))
                header['protocol'] = protocol

                # Read analog-digital conversion sections
                adc = []
                offs = sections['ADC']['index'] * BLOCKSIZE
                size = sections['ADC']['data']
                for i in range(sections['ADC']['length']):
                    ADC = {}
                    f.seek(offs + i * size)
                    for key, form in ABF2_ADC_FIELDS:
                        ADC[key] = ups(read_f(f, form))
                    # Get channel name and unit
                    ADC['ADCChNames'] = (
                        strings[ADC['lADCChannelNameIndex'] - 1])
                    ADC['ADCChUnits'] = strings[ADC['lADCUnitsIndex'] - 1]
                    adc.append(ADC)
                header['listADCInfo'] = adc

                # Read DAC section
                dac = []
                offs = sections['DAC']['index'] * BLOCKSIZE
                size = sections['DAC']['data']
                for i in range(sections['DAC']['length']):
                    f.seek(offs + size * i)
                    DAC = {}
                    for key, form in ABF2_DAC_FIELDS:
                        DAC[key] = ups(read_f(f, form))
                    DAC['sDACChannelName'] = \
                        strings[DAC['lDACChannelNameIndex'] - 1]
                    DAC['sDACChannelUnits'] = \
                        strings[DAC['lDACChannelUnitsIndex'] - 1]
                    dac.append(DAC)
                header['listDACInfo'] = dac

                # Read UserList section
                user_lists = []
                offs = sections['UserList']['index'] * BLOCKSIZE
                size = sections['UserList']['data']
                r = range(sections['UserList']['length'])
                for i in r:  # pragma: no cover
                    # Cover pragma: User lists are not supported
                    f.seek(offs + size * i)
                    user_list = {}
                    for key, form in ABF2_USER_LIST_FIELDS:
                        user_list[key] = ups(read_f(f, form))
                    user_lists.append(user_list)
                header['listUserListInfo'] = user_lists

                # Read epoch-per-DAC section
                # The resulting OrderedDict has the following structure:
                #  - the first index is the DAC number
                #  - the second index is the epoch number
                header['epochInfoPerDAC'] = {}
                offs = sections['EpochPerDAC']['index'] * BLOCKSIZE
                size = sections['EpochPerDAC']['data']
                info = {}
                for i in range(sections['EpochPerDAC']['length']):
                    f.seek(offs + size * i)
                    einf = {}
                    for key, form in ABF2_EPOCH_INFO_PER_DAC_FIELD:
                        einf[key] = ups(read_f(f, form))
                    DACNum = einf['nDACNum']
                    EpochNum = einf['nEpochNum']
                    if DACNum not in info:
                        info[DACNum] = {}
                    info[DACNum][EpochNum] = einf
                header['epochInfoPerDAC'] = info

        self._header = header

    def _read_2_time_of_recording(self):
        """ Read and process the time when this file was recorded. """

        if self._version < 2:
            t1 = str(self._header['lFileStartDate'])
            t2 = float(self._header['lFileStartTime'])
        else:
            t1 = str(self._header['uFileStartDate'])
            t2 = float(self._header['uFileStartTimeMS']) / 1000

        YY = int(t1[0:4])
        MM = int(t1[4:6])
        DD = int(t1[6:8])
        hh = int(t2 / 3600)
        mm = int((t2 - hh * 3600) / 60)
        ss = t2 - hh * 3600 - mm * 60
        ms = int((ss % 1) * 1e6)
        ss = int(ss)

        self._datetime = datetime.datetime(YY, MM, DD, hh, mm, ss, ms)

    def _read_3_protocol_information(self):
        """
        Reads the header fields detailing the number of runs, sweeps, and the
        type of protocol used. Create empty sweeps.
        """
        h = self._header

        # Number of channels, sampling rate (Hz) and acquisition mode
        # Note: Number of A/D channels will be set to 0 if this is a
        #       protocol-only file
        # Note: Number of D/A channels will be adjusted after checking support
        if self._version < 2:
            # In (newer versions of) version 1.x, only 2 D/A channels have
            # full "waveform" support. There are still 4 D/A channels but I
            # don't understand what the other 2 do.
            # 1.x versions only seem to have 1 DAC channel, but this is not
            # supported here.
            self._n_adc = int(h['nADCNumChannels'])
            self._n_dac = min(len(h['sDACChannelName']), 2)
            self._rate = 1e6 / (h['fADCSampleInterval'] * self._n_adc)
            self._mode = h['nOperationMode']
        else:
            # In version 2, there are up to 8 "waveform" D/A channels

            self._n_adc = int(h['sections']['ADC']['length'])
            self._n_dac = int(h['sections']['DAC']['length'])
            self._rate = 1e6 / h['protocol']['fADCSequenceInterval']
            self._mode = h['protocol']['nOperationMode']

        if self._mode not in acquisition_modes:  # pragma: no cover
            raise NotImplementedError(f'Unknown acquisition mode: {mode}')

        # Protocol files don't have A/D channels by definition
        if self._is_protocol_file:
            self._n_adc = 0

        # Only episodic stimulation is supported.
        if self._mode != ACMODE_EPISODIC_STIMULATION:  # pragma: no cover
            warnings.warn(
                'Unsupported acquisition method '
                + acquisition_modes[self._mode] + '; unable to read D/A'
                ' channels.')

            # Remaining code is all about reading D/A info for episodic
            # stimulation, so return
            self._n_dac = 0
            return

        # Gather protocol information
        if self._version < 2:

            # Before version 2: Sections are fixed length, locations absolute
            self._number_of_trials = h['lNumberOfTrials']
            self._trial_start_to_start = h['fTrialStartToStart']
            self._runs_per_trial = h['lRunsPerTrial']
            self._run_start_to_start = h['fRunStartToStart']
            self._sweeps_per_run = h['lSweepsPerRun']
            self._sweep_start_to_start = h['fEpisodeStartToStart']

            # Number of samples in a channel for each sweep
            # (Only works for fixed-length, high-speed-osc or episodic)
            self._samples_per_channel = \
                h['lNumSamplesPerEpisode'] // h['nADCNumChannels']

            def dinfo(index, name):
                """ Return DAC channel info, ABF1 version. """
                return h[name][index]

            def einfo_exists(index):
                """ Check that epoch info exists for a DAC, ABF1 version. """
                # Fields always exist for 2 channels, not always set.
                # But not useful to look at unset ones, so using n_dac instead
                # of hardcoded 2!
                return 0 <= index < self._n_dac

            def einfo(index):
                """ Return epoch info for a DAC, ABF1 version. """
                lo = index * 8
                hi = lo + 8
                for i in range(lo, hi):
                    yield {
                        'type': h['nEpochType'][i],
                        'init_duration': h['lEpochInitDuration'][i],
                        'duration_inc': h['lEpochDurationInc'][i],
                        'init_level': h['fEpochInitLevel'][i],
                        'level_inc': h['fEpochLevelInc'][i],
                    }
            self._epoch_functions = (dinfo, einfo_exists, einfo)

        else:

            # Version 2 uses variable length sections
            p = h['protocol']

            # Trials, runs, sweeps
            # (According to the manual, there should only be 1 trial!)
            self._number_of_trials = p['lNumberOfTrials']
            self._trial_start_to_start = p['fTrialStartToStart']
            self._runs_per_trial = p['lRunsPerTrial']
            self._run_start_to_start = p['fRunStartToStart']
            self._sweeps_per_run = p['lSweepsPerRun']
            self._sweep_start_to_start = p['fSweepStartToStart']

            # Number of samples in a channel in a single sweep
            self._samples_per_channel = \
                p['lNumSamplesPerEpisode'] // h['sections']['ADC']['length']

            # Compatibility functions
            def dinfo(index, name):
                """ Return DAC info, ABF2 version. """
                return h['listDACInfo'][index][name]

            def einfo_exists(index):
                """ Check that epoch info exists for a DAC, ABF2 version. """
                return index in h['epochInfoPerDAC']

            def einfo(index):
                """ Return epoch info for a DAC, ABF2 version. """
                for e in h['epochInfoPerDAC'][index].values():
                    yield {
                        'type': e['nEpochType'],
                        'init_duration': e['lEpochInitDuration'],
                        'duration_inc': e['lEpochDurationInc'],
                        'init_level': e['fEpochInitLevel'],
                        'level_inc': e['fEpochLevelInc'],
                    }
            self._epoch_functions = (dinfo, einfo_exists, einfo)

        # If sweepStartToStart == 0, we set it to the duration of a sweep
        if self._sweep_start_to_start == 0:    # pragma: no cover
            self._sweep_start_to_start = self._samples_per_channel / self._rate

        # Create empty sweeps
        n = h['lActualSweeps']
        if self._is_protocol_file:
            n = self._sweeps_per_run
        self._sweeps = [Sweep() for i in range(n)]

        # User lists are not supported for D/A reconstruction
        # I haven't been able to figure out how you see if a user list is
        # being used, or which channel is using it. There is an 'enable' field
        # but that's been 0 in files that definitely used a UserList...
        # So for now not reading ANY DAC if a userlist even exists.
        user_lists = False
        if self._version < 2:   # pragma: no cover
            user_lists = any(self._header['nULEnable'])
        else:                   # pragma: no cover
            user_lists = len(self._header['listUserListInfo']) > 0
        if user_lists:          # pragma: no cover
            warnings.warn(
                'Unsupported acquisition method: episodic with user lists;'
                ' unable to read D/A channels.')
            self._n_dac = 0
            return

        # Get indices of enabled and supported DAC reconstructions
        supported = {EPOCH_DISABLED, EPOCH_STEPPED}
        for i_dac in range(self._n_dac):
            if einfo_exists(i_dac):
                i = einfo(i_dac)
                use = False

                # Check for unsupported features (or disabled waveforms/epochs)
                # Version 1 files can only have two waveform channels
                if self._version < 2 and i_dac > 1:  # pragma: no cover
                    source = DAC_DISABLED
                else:
                    source = dinfo(i_dac, 'nWaveformSource')
                if source == DAC_EPOCHTABLEWAVEFORM:
                    # Any epoch types besides disabled/stepped? Then don't use
                    # Also don't use if exclusively disabled
                    for e in i:
                        t = e['type']
                        if t == EPOCH_STEPPED:
                            use = True
                        elif t != EPOCH_DISABLED:  # pragma: no cover
                            use = False
                            warnings.warn(
                                f'Unsupported epoch type: {epoch_types(t)}')
                            break
                elif source == DAC_DACFILEWAVEFORM:  # pragma: no cover
                    # Stimulus file? Then don't use
                    warnings.warn('Stimulus file D/A channel not supported.')

                if use:
                    self._dac_indices.append(i_dac)

        # Set true number of D/A outputs
        self._n_dac = len(self._dac_indices)

    def _read_4_ad_conversion_factors(self):
        """ Calculate the factors to convert any integer data to float. """
        self._adc_factors = []
        self._adc_offsets = []
        h = self._header
        if self._version < 2:
            for i in range(self._n_adc):
                j = h['nADCSamplingSeq'][i]

                # Multiplier
                f = (
                    h['fInstrumentScaleFactor'][j]
                    * h['fADCProgrammableGain'][j]
                    * h['lADCResolution']
                    / h['fADCRange'])

                # Signal conditioner used?
                if h['nSignalType'] != 0:   # pragma: no cover
                    # Cover pragma: Don't have appropriate test file
                    f *= h['fSignalGain'][j]

                # Additional gain?
                if h['nTelegraphEnable'][j]:
                    f *= h['fTelegraphAdditGain'][j]

                # Set final gain factor
                self._adc_factors.append(1 / f)

                # Shift
                s = h['fInstrumentOffset'][j]

                # Signal conditioner used?
                if h['nSignalType'] != 0:   # pragma: no cover
                    # Cover pragma: Don't have appropriate test file
                    s -= h['fSignalOffset'][j]

                # Set final offset
                self._adc_offsets.append(s)

        else:

            a = h['listADCInfo']
            p = h['protocol']
            for i in range(self._n_adc):
                # Multiplier
                f = (
                    a[i]['fInstrumentScaleFactor']
                    * a[i]['fADCProgrammableGain']
                    * p['lADCResolution']
                    / p['fADCRange'])

                # Signal conditioner used?
                if h.get('nSignalType', 0) != 0:  # pragma: no cover
                    # Cover pragma: Don't have appropriate test file
                    f *= a[i]['fSignalGain']

                # Additional gain?
                if a[i]['nTelegraphEnable']:
                    f *= a[i]['fTelegraphAdditGain']

                # Set final gain factor
                self._adc_factors.append(1 / f)

                # Shift
                s = a[i]['fInstrumentOffset']

                # Signal conditioner used?
                if h.get('nSignalType', 0) != 0:  # pragma: no cover
                    # Cover pragma: Don't have appropriate test file
                    s -= a[i]['fSignalOffset']

                # Set final offset
                self._adc_offsets.append(s)

    def _read_5_ad_data(self):
        """ Reads the A/D data and appends it to the list of sweeps. """

        h = self._header

        # Sampling rate is constant for all sweeps and channels
        # TODO: This won't work for 2-rate protocols
        rate = self._rate

        # Get binary integer format
        dt = np.dtype('i2') if h['nDataFormat'] == 0 else np.dtype('f4')

        # Get number of channels, create a numpy memory map
        if self._version < 2:
            # Old files, get info from fields stored directly in header
            o = h['lDataSectionPtr'] * BLOCKSIZE \
                + h['nNumPointsIgnored'] * dt.itemsize
            n = h['lActualAcqLength']
        else:
            # New files, get info from appropriate header section
            o = h['sections']['Data']['index'] * BLOCKSIZE
            n = h['sections']['Data']['length']
        data = np.memmap(self._filepath, dt, 'r', shape=(n,), offset=o)

        # Load list of sweeps (Sweeps are called 'episodes' in ABF < 2)
        if self._version < 2:
            n = h['lSynchArraySize']
            o = h['lSynchArrayPtr'] * BLOCKSIZE
        else:
            n = h['sections']['SynchArray']['length']
            o = h['sections']['SynchArray']['index'] * BLOCKSIZE
        if n > 0:
            dt = [(str('offset'), str('i4')), (str('len'), str('i4'))]
            sweep_data = np.memmap(
                self._filepath, dt, 'r', shape=(n,), offset=o)
        else:   # pragma: no cover
            # Cover pragma: Don't have appropriate test file
            sweep_data = np.empty((1), dt)
            sweep_data[0]['len'] = data.size
            sweep_data[0]['offset'] = 0

        # Number of sweeps must equal n
        if n != h['lActualSweeps']:
            raise NotImplementedError(
                'Unable to read file with different sizes per sweep.')

        # Time-offset at start of first sweep
        start = sweep_data[0]['offset'] / rate

        # Get data
        pos = 0
        for i_sweep, sdat in enumerate(sweep_data):

            # Get the number of data points
            size = sdat['len']

            # Calculate the correct size for variable-length event mode
            if self._mode == ACMODE_VARIABLE_LENGTH_EVENTS:  # pragma: no cover
                # Cover pragma: Only episodic stimulus is supported.
                if self._version < 2:
                    f = float(h['fSynchTimeUnit'])
                else:
                    f = float(h['protocol']['fSynchTimeUnit'])
                if f != 0:
                    size /= f

            # Get a memory map to the relevant part of the data
            part = data[pos: pos + size]
            pos += size
            part = part.reshape(
                (part.size // self._n_adc, self._n_adc)).astype('f')

            # If needed, reformat the integers
            if h['nDataFormat'] == 0:
                # Data given as integers? Convert to floating point

                for i in range(self._n_adc):
                    part[:, i] *= self._adc_factors[i]
                    part[:, i] += self._adc_offsets[i]

            # Get start in other modes
            if self._mode != ACMODE_EPISODIC_STIMULATION:  # pragma: no cover
                # All modes except episodic stimulation
                start = data['offset'] / rate

            # Create and populate sweep
            sweep = self._sweeps[i_sweep]
            for i in range(self._n_adc):
                c = Channel(self)
                c._data = part[:, i]    # Actually store the data
                c._rate = rate
                c._start = start

                if self._version < 2:
                    j = h['nADCSamplingSeq'][i]

                    c._name = h['sADCChannelName'][j]
                    c._index = int(h['nADCPtoLChannelMap'][j])
                    c._unit = self._unit(h['sADCUnits'][j])

                    # Get telegraphed info
                    def get(field):
                        try:
                            return float(h[field][j])
                        except KeyError:
                            return None

                    if get('nTelegraphEnable'):
                        c._type = int(get('nTelegraphMode') or 0)
                        c._cm = get('fTelegraphMembraneCap')
                        c._rs = get('fTelegraphAccessResistance')
                        c._lopass = get('fTelegraphFilter')

                    # Updated low-pass cutoff
                    if h['nSignalType'] != 0:  # pragma: no cover
                        # Cover pragma: Don't have appropriate test file
                        # If a signal conditioner is used, the cutoff frequency
                        # is an undescribed "complex function" of both low-pass
                        # settings...
                        c._lopass = None

                else:
                    c._name = h['listADCInfo'][i]['ADCChNames']
                    c._index = int(h['listADCInfo'][i]['nADCNum'])
                    c._unit = self._unit(h['listADCInfo'][i]['ADCChUnits'])

                    # Get telegraphed info
                    if h['listADCInfo'][i]['nTelegraphEnable']:
                        c._type = int(h['listADCInfo'][i]['nTelegraphMode'])
                        c._cm = float(
                            h['listADCInfo'][i]['fTelegraphMembraneCap'])
                        c._rs = float(
                            h['listADCInfo'][i]['fTelegraphAccessResistance'])
                        c._lopass = float(
                            h['listADCInfo'][i]['fTelegraphFilter'])

                    # Updated low-pass cutoff
                    if 'nSignalType' in h['protocol']:  # pragma: no cover
                        # Cover pragma: Don't have appropriate test file
                        if h['protocol']['nSignalType'] != 0:
                            # If a signal conditioner is used, the cutoff
                            # frequency is an undescribed "complex function" of
                            # both low-pass settings...
                            c._lopass = None

                sweep._channels.append(c)

            if self._mode == ACMODE_EPISODIC_STIMULATION:
                # Increase time according to sweeps in episodic stim. mode
                start += self._sweep_start_to_start

    def _read_6_da_reconstructions(self):
        """
        Convert supported D/A waveforms to channels.

        Only works for episodic stimulation, with step protocols and no
        user lists.

        The resulting analog signal has the same size as the recorded
        signals, so not always the full length of the protocol!

        """
        dinfo, einfo_exists, einfo = self._epoch_functions

        ns = self._samples_per_channel
        start = 0
        for i_sweep, sweep in enumerate(self._sweeps):
            for i_dac in self._dac_indices:

                # Create a channel
                c = Channel(self)
                c._name = dinfo(i_dac, 'sDACChannelName')
                if self._version < 2:
                    c._index = i_dac
                else:
                    c._index = int(dinfo(i_dac, 'lDACChannelNameIndex'))
                c._data = np.ones(ns) * dinfo(i_dac, 'fDACHoldingLevel')
                c._rate = self._rate
                c._start = start
                c._unit = self._unit(dinfo(i_dac, 'sDACChannelUnits'))

                # Find start of first epoch. This is defined as being at t=0
                # but axon likes to add some samples before the first and after
                # the last epoch. We can find out the number of samples using
                # a procedure found in ABF v1's _GetHoldingLength()
                if self._is_protocol_file:
                    i2 = 0
                else:
                    i2 = ns // 64  # ABFH_HOLDINGFRACTION = 64
                    i2 -= i2 % self._n_adc
                    if (i2 < self._n_adc):  # pragma: no cover
                        i2 = self._n_adc

                # For each 'epoch' in the stimulation signal
                for e in einfo(i_dac):
                    if e['type'] == EPOCH_STEPPED:
                        dur = e['init_duration']
                        inc = e['duration_inc']
                        i1 = i2
                        i2 += dur + i_sweep * inc
                        level = e['init_level'] + e['level_inc'] * i_sweep
                        c._data[i1:i2] = level * np.ones(len(range(i2 - i1)))

                # Store channel
                sweep._channels.append(c)

            # Update start for next sweep
            start += self._sweep_start_to_start

    def __getitem__(self, key):
        return self._sweeps[key]

    def __iter__(self):
        return iter(self._sweeps)

    def __len__(self):
        return len(self._sweeps)

    def _channel_id(self, channel_id):
        """ Checks an int or str channel id and returns a valid int. """
        if len(self._sweeps) == 0:  # pragma: no cover
            raise KeyError(f'Channel {channel_id} not found (empty file).')

        # Handle string
        if isinstance(channel_id, str):
            int_id = self._ad_names[channel_id]  # Bubble KeyError to user
        else:
            int_id = int(channel_id)    # Propagate TypeError
            if int_id < 0 or int_id >= self._n_adc:
                raise IndexError(f'channel_id out of range: {channel_id}')

        return int_id

    def channel(self, channel_id, join_sweeps=False):
        # Docstring in SweepSource
        channel_id = self._channel_id(channel_id)
        time, data = [], []
        for i, sweep in enumerate(self._sweeps):
            time.append(sweep[channel_id].times())
            data.append(sweep[channel_id].values())
        if join_sweeps:
            return (np.concatenate(time), np.concatenate(data))
        return time, data

    def channel_count(self):
        # Docstring in SweepSource
        return self._n_adc

    def channel_names(self, index=None):
        # Docstring in SweepSource
        if index is None:
            return list(self._ad_names.keys())
        return list(self._ad_names.keys())[index]

    def channel_units(self, index=None):
        # Docstring in SweepSource
        if index is None:
            return list(self._ad_units)
        return self._ad_units[index]

    def _da_id(self, output_id):
        """
        Checks an int or str D/A channel id and returns a valid int.

        Note: The integer here is from 0 to da_count(), so not equal to the
        channel :meth:`index()` shown in pclamp.
        """
        if len(self._sweeps) == 0:  # pragma: no cover
            raise KeyError(f'D/A output {output_id} not found (empty file).')

        # Handle string
        if isinstance(output_id, str):
            int_id = self._da_names[output_id]  # Propagate KeyError to user
        else:
            int_id = int(output_id)  # Propagate TypeError
            if int_id < 0 or int_id >= self._n_dac:
                raise IndexError(f'output_id out of range: {output_id}')

        return int_id

    def da(self, output_id, join_sweeps=False):
        # Docstring in SweepSource
        channel_id = self._n_adc + self._da_id(output_id)
        time, data = [], []
        for i, sweep in enumerate(self._sweeps):
            time.append(sweep[channel_id].times())
            data.append(sweep[channel_id].values())
        if join_sweeps:
            return (np.concatenate(time), np.concatenate(data))
        return time, data

    def da_count(self):
        # Docstring in SweepSource
        return self._n_dac

    def da_names(self, index=None):
        # Docstring in SweepSource
        if index is None:
            return list(self._da_names.keys())
        return list(self._da_names.keys())[index]

    def da_protocol(self, output_id=None, tu='ms', vu='mV', cu='pA',
                    n_digits=9, include_initial_holding=False):
        """
        See :meth:`myokit.formats.SweepSource.da_protocol()`.

        This implementation adds a keyword argument ``include_initial_holding``
        that lets you switch between the declared protocol (``False``) and the
        protocol as actually implemented (``True``). In the latter case, a
        short holding time is added before the first epoch in every sweep.
        """

        # Check the output id. This also raises an error if no supported D/A
        # channels are present.
        output_id = self._da_id(output_id or 0)

        # Get the index in dinfo
        i_dac = self._dac_indices[output_id]
        dinfo, einfo_exists, einfo = self._epoch_functions

        # Get the time and data conversion factors
        units = myokit.units
        tf = myokit.Unit.conversion_factor(units.s, tu)
        if myokit.Unit.can_convert(self._da_units[output_id], units.V):
            df = myokit.Unit.conversion_factor(self._da_units[output_id], vu)
        elif myokit.Unit.can_convert(
                self._da_units[output_id], units.A):  # pragma: no cover
            df = myokit.Unit.conversion_factor(self._da_units[output_id], cu)
        else:  # pragma: no cover
            # Not a voltage or current? Then don't convert
            df = 1
        tf, df = float(tf), float(df)

        # Axon has the annoying habit of adding some extra holding at the start
        # We can include this if we want. See _read_6 for details.
        if self._is_protocol_file:
            offset = 0
        else:
            offset = self._samples_per_channel // 64
            offset -= offset % self._n_adc
            if (offset < self._n_adc):  # pragma: no cover
                # Don't have a test for this, but this is part of the
                # established procedure.
                offset = self._n_adc
            offset /= self._rate

        # Holding level (converted and rounded)
        holding = round(df * dinfo(i_dac, 'fDACHoldingLevel'), n_digits)

        # Create protocol
        p = myokit.Protocol()
        start = 0
        next_start = self._sweep_start_to_start
        for i_sweep in range(self._sweeps_per_run):
            # Start of sweep: secret event at holding potential
            if include_initial_holding:
                e_start = round(tf * start, n_digits)
                e_length = round(tf * offset, n_digits)
                p.schedule(holding, e_start, e_length)
                start += offset

            for e in einfo(i_dac):
                if e['type'] == EPOCH_STEPPED:
                    dur = e['init_duration'] / self._rate
                    inc = e['duration_inc'] / self._rate
                    duration = dur + i_sweep * inc
                    level = e['init_level'] + e['level_inc'] * i_sweep

                    e_level = round(df * level, n_digits)
                    e_start = round(tf * start, n_digits)
                    e_length = round(tf * duration, n_digits)
                    p.schedule(e_level, e_start, e_length)

                    start += duration
                # Note: Only other type can be EPOCH_DISABLED at this point

            # End of sweep: event at holding potential
            e_start = round(tf * start, n_digits)
            e_length = round(tf * (next_start - start), n_digits)
            p.schedule(holding, e_start, e_length)
            start = next_start
            next_start += self._sweep_start_to_start

        return p

    def da_units(self, index=None):
        # Docstring in SweepSource
        if index is None:
            return list(self._da_units)
        return self._da_units[index]

    def equal_length_sweeps(self):
        # Always true for ABF
        return True

    def filename(self):
        """ Returns this ABF file's filename. """
        return self._filename

    def log(self, join_sweeps=False, use_names=False, include_da=True):
        # Docstring in SweepSource

        # Create log, return if no sweeps or channels
        log = myokit.DataLog()
        ns = len(self._sweeps)
        if ns == 0 or (self._n_adc + self._n_dac) == 0:  # pragma: no cover
            return log

        # Get channel names
        if use_names:
            nc = self._n_adc + (self._n_dac if include_da else 0)
            names = [c.name() for c in self._sweeps[0][:nc]]
        else:
            names = [f'{i}.channel' for i in range(self._n_adc)]
            if include_da:
                names += [f'{i}.da' for i in range(self._n_dac)]

        # Gather data and return
        t = self._sweeps[0][0].times()
        if not join_sweeps:
            log['time'] = t
            for i_sweep, sweep in enumerate(self._sweeps):
                for channel, name in zip(sweep, names):
                    log[name, i_sweep] = channel.values()
        else:
            log['time'] = np.concatenate(
                [t + i * self._sweep_start_to_start for i in range(ns)])
            for i_channel, name in enumerate(names):
                log[name] = np.concatenate(
                    [sweep[i_channel].values() for sweep in self._sweeps])
        log.set_time_key('time')
        return log

    def matplotlib_figure(self):
        """ Creates and returns a matplotlib figure with this file's data. """
        import matplotlib.pyplot as plt
        f = plt.figure()
        plt.suptitle(self.filename())

        # Plot AD channels
        ax = plt.subplot(2, 1, 1)
        ax.set_title('Measured data')
        times = None
        for sweep in self._sweeps:
            for channel in sweep[:self._n_adc]:
                if times is None:
                    times = channel.times()
                plt.plot(times, channel.values())

        # Plot DA channels
        n = self._n_dac
        ax = [plt.subplot(2, n, n + 1 + i) for i in range(n)]
        for sweep in self._sweeps:
            for i, channel in enumerate(sweep[self._n_adc:]):
                ax[i].set_title(channel.name())
                ax[i].plot(times, channel.values())

        return f

    def meta_str(self, show_header=False):
        """
        Returns a multi-line string with meta data about this file.

        The optional argument ``show_header`` can be used to add the full
        header contents to the output.
        """
        out = []

        # File info
        if self._is_protocol_file:
            out.append(f'Axon Protocol File: {self._filename}')
        else:
            out.append(f'Axon Binary File: {self._filename}')
        out.append(f'ABF Format version {self._version_str}')
        out.append(f'Recorded on: {self._datetime}')

        # AProtocol info
        out.append(
            f'Acquisition mode: {self._mode}: {acquisition_modes[self._mode]}')
        if self._number_of_trials:
            out.append(
                f'Protocol set for {self._number_of_trials} trials,'
                f' spaced {self._trial_start_to_start}s apart.')
            out.append(
                f'    with {self._runs_per_trial} runs per trial,'
                f' spaced {self._run_start_to_start}s apart.')
            out.append(
                f'     and {self._sweeps_per_run} sweeps per run,'
                f' spaced {self._sweep_start_to_start}s apart.')
        else:   # pragma: no cover
            out.append('Protocol data could not be determined.')
        out.append(f'Sampling rate: {self._rate} Hz')

        # Channel info
        if len(self._sweeps) > 0:

            # A/D recordings
            for i, c in enumerate(self._sweeps[0][:self._n_adc]):
                out.append(f'A/D Channel {i}: "{c._name}"')
                if c._type:  # pragma: no cover
                    # Cover pragma: Don't have appropriate test file
                    out.append(f'  Type: {type_mode_names[c._type]}')
                out.append(f'  Unit: {c._unit}')
                if c._lopass:
                    out.append(f'  Low-pass filter: {c._lopass} Hz')
                if c._cm:
                    out.append(f'  Cm (telegraphed): {c._cm} pF')
                if c._rs:   # pragma: no cover
                    # Cover pragma: Don't have appropriate test file
                    out.append(f'  Rs (telegraphed): {c._rs}')

            # Reconstructed D/A outputs
            for i, c in enumerate(self._sweeps[0][self._n_adc:]):
                out.append(f'D/A Channel {i}: "{c._name}"')
                if c._type:  # pragma: no cover
                    # Cover pragma: Don't have appropriate test file
                    out.append('  Type: {type_mode_names[c._type]}')
                out.append(f'  Unit: {c._unit}')

        # Add full header info
        if show_header:
            if self._strings:
                dict_to_string(out, 'Strings', {'strings': self._strings})
            dict_to_string(out, 'file header', self._header)

        return '\n'.join(out)

    def path(self):
        """ Returns the path to the underlying ABF file. """
        return self._filepath

    def sweep_count(self):
        # Docstring in SweepSource
        return len(self._sweeps)

    def time_unit(self):
        # Docstring in SweepSource
        # For ABF, this is always seconds
        return myokit.units.s

    def _unit(self, unit_string):
        """ Parses a unit string and returns a :class:`myokit.Unit`. """
        try:
            return self._unit_cache[unit_string]
        except KeyError:
            unit = myokit.parse_unit(unit_string.replace(MU, 'u'))
            self._unit_cache[unit_string] = unit
            return unit

    def version(self):
        """ Returns a string representation of this file's version number. """
        return self._version_str


class Sweep:
    """
    Represents a single sweep (also called an *episode*).

    Each sweep contains a fixed number of :class:`channels<Channel>`.
    """
    def __init__(self):
        self._channels = []

    def __getitem__(self, key):
        return self._channels[key]  # Handles slices etc.

    def __iter__(self):
        return iter(self._channels)

    def __len__(self):
        return len(self._channels)


class Channel:
    """
    Represents a signal for a single channel.

    To obtain its data, use :meth:`times` and :meth:`values`.
    """
    def __init__(self, parent_file):
        self._parent_file = parent_file  # The abf file this channel is from
        self._type = TYPE_UNKNOWN   # Type of recording

        # This channel's name
        self._name = None

        # This channel's index in the file. This is basically a name, and does
        # not correspond to e.g. its index in the ADC/DAC info or its index in
        # the sweep's list of channels.
        self._index = None

        # The units this channel's data is in
        self._unit = None

        # The raw data points
        self._data = None

        # Sampling rate in Hz
        self._rate = None

        # The signal start time
        self._start = None

        # The reported membrane capacitance
        self._cm = None

        # The reported access resistance
        self._rs = None

        # The reported low-pass filter cut-off frequency
        self._lopass = None

    def index(self):
        """ Returns the index set for this channel. """
        return self._index

    def name(self):
        """ Returns the name set for this channel. """
        return self._name

    def __str__(self):
        return (
            f'Channel({self._index} "{self._name}"); {len(self._data)} points'
            f' sampled at {self._rate}Hz, starts at t={self._start}.')

    def times(self):
        """ Returns a copy of the values on the time axis. """
        n = len(self._data)
        f = 1 / self._rate
        return np.arange(self._start, self._start + n * f, f)[0:n]

    def unit(self):
        """ Returns the units this channel is in. """
        return self._unit

    def values(self):
        """ Returns a copy of the values on the data axis. """
        return np.copy(self._data)


def dict_to_string(out, name, d, tab=''):
    """ Used by AbfFile.info(). """
    m = max(0, 38 - len(tab) - int(0.1 + len(name) / 2))
    out.append(f'{tab}{"-" * m}  {name}  {"-" * m}')
    for n, v in d.items():
        n = str(n)
        if type(v) == dict:
            dict_to_string(out, n, v, f'{tab}  ')
        elif type(v) == list:
            list_to_string(out, n, v, tab)
        else:
            out.append(f'{tab}{n}: {v}')
    m = max(0, 80 - 2 * len(tab))
    out.append(f'{tab}{m * "-"}')


def list_to_string(out, name, d, tab=''):
    """ Used by AbfFile.info(). """
    for index, item in enumerate(d):
        n = f'{name}[{index}]'
        if type(item) == dict:
            dict_to_string(out, n, item, tab)
        elif type(item) == list:    # pragma: no cover
            # Cover pragma: Don't have appropriate test file
            list_to_string(out, n, item, tab)
        else:
            out.append(f'{tab}{n}: {item}')


# Some python struct types:
#   f   float
#   h   short
#   i   int
#   s   string
# Size of block alignment in ABF Files
BLOCKSIZE = 512


# A mu, sometimes found in unit strings
MU = '\u00b5'


# Header fields for versions 1 and 2
# Stored as (key, offset, format) where format corresponds to a struct
#  unpacking format as documented in:
#  http://docs.python.org/library/struct.html#format-characters
HEADER_FIELDS = {
    # Note that a lot of the groups in the version 1 header start with obsolete
    # fields, followed later by their newer equivalents.
    1: [
        ('fFileSignature', 0, '4s'),       # Coarse file version indication
        # Group 1, File info and sizes
        ('fFileVersionNumber', 4, 'f'),    # Version number as float
        ('nOperationMode', 8, 'h'),        # Acquisition mode
        ('lActualAcqLength', 10, 'i'),
        ('nNumPointsIgnored', 14, 'h'),
        ('lActualSweeps', 16, 'i'),
        ('lFileStartDate', 20, 'i'),
        ('lFileStartTime', 24, 'i'),
        ('lStopWatchTime', 28, 'i'),
        ('fHeaderVersionNumber', 32, 'f'),
        ('nFileType', 36, 'h'),
        ('nMSBinFormat', 38, 'h'),
        # Group 2, file structure
        ('lDataSectionPtr', 40, 'i'),
        ('lTagSectionPtr', 44, 'i'),
        ('lNumTagEntries', 48, 'i'),
        ('lScopeConfigPtr', 52, 'i'),
        ('lNumScopes', 56, 'i'),
        ('x_lDACFilePtr', 60, 'i'),
        ('x_lDACFileNumEpisodes', 64, 'i'),
        ('lDeltaArrayPtr', 72, 'i'),
        ('lNumDeltas', 76, 'i'),
        ('lVoiceTagPtr', 80, 'i'),
        ('lVoiceTagEntries', 84, 'i'),
        ('lSynchArrayPtr', 92, 'i'),
        ('lSynchArraySize', 96, 'i'),
        ('nDataFormat', 100, 'h'),
        ('nSimultaneousScan', 102, 'h'),
        # Group 3, Trial hierarchy
        ('nADCNumChannels', 120, 'h'),
        ('fADCSampleInterval', 122, 'f'),
        ('fADCSecondSampleInterval', 126, 'f'),
        ('fSynchTimeUnit', 130, 'f'),
        ('fSecondsPerRun', 134, 'f'),
        ('lNumSamplesPerEpisode', 138, 'i'),
        ('lPreTriggerSamples', 142, 'i'),
        ('lSweepsPerRun', 146, 'i'),        # Number of sweeps/episodes per run
        ('lRunsPerTrial', 150, 'i'),
        ('lNumberOfTrials', 154, 'i'),
        ('nAveragingMode', 158, 'h'),
        ('nUndoRunCount', 160, 'h'),
        ('nFirstEpisodeInRun', 162, 'h'),
        ('fTriggerThreshold', 164, 'f'),
        ('nTriggerSource', 168, 'h'),
        ('nTriggerAction', 170, 'h'),
        ('nTriggerPolarity', 172, 'h'),
        ('fScopeOutputInterval', 174, 'f'),
        ('fEpisodeStartToStart', 178, 'f'),
        ('fRunStartToStart', 182, 'f'),
        ('fTrialStartToStart', 186, 'f'),
        ('lAverageCount', 190, 'f'),
        ('lClockChange', 194, 'f'),
        ('nAutoTriggerStrategy', 198, 'h'),
        # Group 4, Display parameters
        # Group 5, Hardware info
        ('fADCRange', 244, 'f'),
        ('lADCResolution', 252, 'i'),
        # Group 6, Environment info
        ('nFileStartMillisecs', 366, 'h'),
        # Group 7, Multi-channel info
        ('nADCPtoLChannelMap', 378, '16h'),
        ('nADCSamplingSeq', 410, '16h'),
        ('sADCChannelName', 442, '10s' * 16),
        ('sADCUnits', 602, '8s' * 16),
        ('fADCProgrammableGain', 730, '16f'),
        ('fADCDisplayAmplification', 794, '16f'),
        ('fADCDisplayOffset', 858, '16f'),
        ('fInstrumentScaleFactor', 922, '16f'),
        ('fInstrumentOffset', 986, '16f'),
        # The fSignal fields are only relevant if a signal conditioner was used
        ('fSignalGain', 1050, '16f'),
        ('fSignalOffset', 1114, '16f'),
        ('fSignalLowpassFilter', 1178, '16f'),
        ('fSignalHighpassFilter', 1242, '16f'),
        ('sDACChannelName', 1306, '10s' * 4),
        ('sDACChannelUnits', 1346, '8s' * 4),
        ('fDACScaleFactor', 1378, '4f'),
        ('fDACHoldingLevel', 1394, '4f'),
        # 1 if a "CyberAmp 320/380" signal conditioner was used
        ('nSignalType', 1410, 'h'),
        # Group 8, There doesn't seem to be a group 8
        # Group 9, Wave data
        ('nDigitalEnable', 1436, 'h'),
        ('x_nWaveformSource', 1438, 'h'),
        ('nActiveDACChannel', 1440, 'h'),
        ('x_nInterEpisodeLevel', 1442, 'h'),
        ('x_nEpochType', 1444, '10h'),
        ('x_fEpochInitLevel', 1464, '10f'),
        ('x_fEpochLevelInc', 1504, '10f'),
        ('x_nEpochInitDuration', 1544, '10h'),
        ('x_nEpochDurationInc', 1564, '10h'),
        ('nDigitalHolding', 1584, 'h'),
        ('nDigitalInterEpisode', 1586, 'h'),
        ('nDigitalValue', 2588, '10h'),
        ('lDACFilePtr', 2048, '2i'),
        ('lDACFileNumEpisodes', 2056, '2i'),
        ('fDACCalibrationFactor', 2074, '4f'),
        ('fDACCalibrationOffset', 2090, '4f'),
        ('nWaveformEnable', 2296, '2h'),
        ('nWaveformSource', 2300, '2h'),
        ('nInterEpisodeLevel', 2304, '2h'),
        ('nEpochType', 2308, '20h'),       # 2 CMD channels with 10 values each
        ('fEpochInitLevel', 2348, '20f'),
        ('fEpochLevelInc', 2428, '20f'),
        ('lEpochInitDuration', 2508, '20i'),
        ('lEpochDurationInc', 2588, '20i'),
        # Group 10, DAC Output file (Stimulus file)
        ('fDACFileScale', 2708, 'd'),
        ('fDACFileOffset', 2716, 'd'),
        ('lDACFileEpisodeNum', 2724, 'i'),
        ('nDACFileADCNum', 2732, '2h'),
        ('sDACFilePath', 2736, '256s' * 2),     # Two strings
        # Group 11,
        # Group 12, User list parameters
        ('nULEnable', 3360, '4h'),
        ('nULParamToVary', 3368, '4h'),
        ('nULParamValueList0', 3376, '256s' * 4),
        ('nULRepeat', 4400, '4h'),
        # Group 13,
        # Group 14,
        # Group 15, Leak subtraction
        # Group 16, Misc
        # Group 17, Trains
        # Group 18, Application version data
        # Group 19
        # Group 20
        # Group 21 Skipped
        # Group 22
        # Group 23 Post-processing
        # Group 24 Legacy stuff
        # Group 6 extended
        ('nTelegraphEnable', 4512, '16h'),
        ('fTelegraphAdditGain', 4576, '16f'),
    ],
    2: [
        ('fFileSignature', 0, '4s'),       # Coarse file version indication
        ('fFileVersionNumber', 4, '4b'),   # Version number as 4 signed chars
        ('uFileInfoSize', 8, 'I'),
        ('lActualSweeps', 12, 'I'),
        ('uFileStartDate', 16, 'I'),       # File start data YYYYMMDD
        ('uFileStartTimeMS', 20, 'I'),     # Time of day in ms ?
        ('uStopwatchTime', 24, 'I'),
        ('nFileType', 28, 'H'),
        ('nDataFormat', 30, 'H'),
        ('nSimultaneousScan', 32, 'H'),
        ('nCRCEnable', 34, 'H'),
        ('uFileCRC', 36, 'I'),
        ('FileGUID', 40, 'I'),
        ('uCreatorVersion', 56, 'I'),
        ('uCreatorNameIndex', 60, 'I'),
        ('uModifierVersion', 64, 'I'),
        ('uModifierNameIndex', 68, 'I'),
        ('uProtocolPathIndex', 72, 'I')
    ]
}


# ABF2 File sections
ABF2_FILE_SECTIONS = [
    'Protocol',
    'ADC',
    'DAC',
    'Epoch',
    'ADCPerDAC',
    'EpochPerDAC',
    'UserList',
    'StatsRegion',
    'Math',
    'Strings',
    'Data',
    'Tag',
    'Scope',
    'Delta',
    'VoiceTag',
    'SynchArray',
    'Annotation',
    'Stats',
]


# ABF2 Fields in the tag section
ABF2_TAG_INFO_DESCRIPTION = [
    ('lTagTime', 'i'),
    ('sComment', '56s'),
    ('nTagType', 'h'),
    ('nVoiceTagNumber_or_AnnotationIndex', 'h'),
]


# ABF2 Fields in the protocol section
ABF2_PROTOCOL_FIELDS = [
    ('nOperationMode', 'h'),                # 0
    ('fADCSequenceInterval', 'f'),          # 2
    ('bEnableFileCompression', 'b'),        # 6
    ('sUnused1', '3s'),                     # 7
    ('uFileCompressionRatio', 'I'),         # 10
    ('fSynchTimeUnit', 'f'),                # 14
    ('fSecondsPerRun', 'f'),                # 18
    ('lNumSamplesPerEpisode', 'i'),         # 22
    ('lPreTriggerSamples', 'i'),            # 26
    ('lSweepsPerRun', 'i'),                 # 30
    ('lRunsPerTrial', 'i'),                 # 34
    ('lNumberOfTrials', 'i'),               # 38
    ('nAveragingMode', 'h'),                # 42
    ('nUndoRunCount', 'h'),                 # 44
    ('nFirstEpisodeInRun', 'h'),            # 46
    ('fTriggerThreshold', 'f'),             # 48
    ('nTriggerSource', 'h'),                # 52
    ('nTriggerAction', 'h'),                # 54
    ('nTriggerPolarity', 'h'),              # 56
    ('fScopeOutputInterval', 'f'),          # 58
    ('fSweepStartToStart', 'f'),            # 62
    ('fRunStartToStart', 'f'),
    ('lAverageCount', 'i'),
    ('fTrialStartToStart', 'f'),
    ('nAutoTriggerStrategy', 'h'),
    ('fFirstRunDelayS', 'f'),
    ('nChannelStatsStrategy', 'h'),
    ('lSamplesPerTrace', 'i'),
    ('lStartDisplayNum', 'i'),
    ('lFinishDisplayNum', 'i'),
    ('nShowPNRawData', 'h'),
    ('fStatisticsPeriod', 'f'),
    ('lStatisticsMeasurements', 'i'),
    ('nStatisticsSaveStrategy', 'h'),
    ('fADCRange', 'f'),
    ('fDACRange', 'f'),
    ('lADCResolution', 'i'),
    ('lDACResolution', 'i'),
    ('nExperimentType', 'h'),
    ('nManualInfoStrategy', 'h'),
    ('nCommentsEnable', 'h'),
    ('lFileCommentIndex', 'i'),
    ('nAutoAnalyseEnable', 'h'),
    ('nSignalType', 'h'),
    ('nDigitalEnable', 'h'),
    ('nActiveDACChannel', 'h'),
    ('nDigitalHolding', 'h'),
    ('nDigitalInterEpisode', 'h'),
    ('nDigitalDACChannel', 'h'),
    ('nDigitalTrainActiveLogic', 'h'),
    ('nStatsEnable', 'h'),
    ('nStatisticsClearStrategy', 'h'),
    ('nLevelHysteresis', 'h'),
    ('lTimeHysteresis', 'i'),
    ('nAllowExternalTags', 'h'),
    ('nAverageAlgorithm', 'h'),
    ('fAverageWeighting', 'f'),
    ('nUndoPromptStrategy', 'h'),
    ('nTrialTriggerSource', 'h'),
    ('nStatisticsDisplayStrategy', 'h'),
    ('nExternalTagType', 'h'),
    ('nScopeTriggerOut', 'h'),
    ('nLTPType', 'h'),
    ('nAlternateDACOutputState', 'h'),
    ('nAlternateDigitalOutputState', 'h'),
    ('fCellID', '3f'),
    ('nDigitizerADCs', 'h'),
    ('nDigitizerDACs', 'h'),
    ('nDigitizerTotalDigitalOuts', 'h'),
    ('nDigitizerSynchDigitalOuts', 'h'),
    ('nDigitizerType', 'h'),
]


# ABF2 Fields in the ADC section
ABF2_ADC_FIELDS = [
    ('nADCNum', 'h'),
    ('nTelegraphEnable', 'h'),
    ('nTelegraphInstrument', 'h'),
    ('fTelegraphAdditGain', 'f'),
    ('fTelegraphFilter', 'f'),
    ('fTelegraphMembraneCap', 'f'),
    ('nTelegraphMode', 'h'),
    ('fTelegraphAccessResistance', 'f'),
    ('nADCPtoLChannelMap', 'h'),
    ('nADCSamplingSeq', 'h'),
    ('fADCProgrammableGain', 'f'),
    ('fADCDisplayAmplification', 'f'),
    ('fADCDisplayOffset', 'f'),
    ('fInstrumentScaleFactor', 'f'),
    ('fInstrumentOffset', 'f'),
    ('fSignalGain', 'f'),      # The fSignal fields are only relevant if a
    ('fSignalOffset', 'f'),    # signal conditioner was used
    ('fSignalLowpassFilter', 'f'),
    ('fSignalHighpassFilter', 'f'),
    ('nLowpassFilterType', 'b'),
    ('nHighpassFilterType', 'b'),
    ('fPostProcessLowpassFilter', 'f'),
    ('nPostProcessLowpassFilterType', 'c'),
    ('bEnabledDuringPN', 'b'),
    ('nStatsChannelPolarity', 'h'),
    ('lADCChannelNameIndex', 'i'),
    ('lADCUnitsIndex', 'i'),
]


# ABF2 Fields in the DAC section
ABF2_DAC_FIELDS = [
    ('nDACNum', 'h'),
    ('nTelegraphDACScaleFactorEnable', 'h'),
    ('fInstrumentHoldingLevel', 'f'),
    ('fDACScaleFactor', 'f'),
    ('fDACHoldingLevel', 'f'),
    ('fDACCalibrationFactor', 'f'),
    ('fDACCalibrationOffset', 'f'),
    ('lDACChannelNameIndex', 'i'),
    ('lDACChannelUnitsIndex', 'i'),
    ('lDACFilePtr', 'i'),
    ('lDACFileNumSweeps', 'i'),
    ('nWaveformEnable', 'h'),
    ('nWaveformSource', 'h'),
    ('nInterEpisodeLevel', 'h'),
    ('fDACFileScale', 'f'),
    ('fDACFileOffset', 'f'),
    ('lDACFileEpisodeNum', 'i'),
    ('nDACFileADCNum', 'h'),
    ('nConditEnable', 'h'),
    ('lConditNumPulses', 'i'),
    ('fBaselineDuration', 'f'),
    ('fBaselineLevel', 'f'),
    ('fStepDuration', 'f'),
    ('fStepLevel', 'f'),
    ('fPostTrainPeriod', 'f'),
    ('fPostTrainLevel', 'f'),
    ('nMembTestEnable', 'h'),
    ('nLeakSubtractType', 'h'),
    ('nPNPolarity', 'h'),
    ('fPNHoldingLevel', 'f'),
    ('nPNNumADCChannels', 'h'),
    ('nPNPosition', 'h'),
    ('nPNNumPulses', 'h'),
    ('fPNSettlingTime', 'f'),
    ('fPNInterpulse', 'f'),
    ('nLTPUsageOfDAC', 'h'),
    ('nLTPPresynapticPulses', 'h'),
    ('lDACFilePathIndex', 'i'),
    ('fMembTestPreSettlingTimeMS', 'f'),
    ('fMembTestPostSettlingTimeMS', 'f'),
    ('nLeakSubtractADCIndex', 'h'),
    ('sUnused', '124s'),
]


# ABF2 Fields in the DAC-Epoch section
ABF2_EPOCH_INFO_PER_DAC_FIELD = [
    ('nEpochNum', 'h'),
    ('nDACNum', 'h'),
    ('nEpochType', 'h'),
    ('fEpochInitLevel', 'f'),
    ('fEpochLevelInc', 'f'),
    ('lEpochInitDuration', 'i'),
    ('lEpochDurationInc', 'i'),
    ('lEpochPulsePeriod', 'i'),
    ('lEpochPulseWidth', 'i'),
    ('sUnused', '18s'),
]

# ABF2 User list fields
ABF2_USER_LIST_FIELDS = [
    ('nListNum', 'h'),
    ('nULEnable', 'h'),
    ('nULParamToVary', 'h'),
    ('nULRepeat', 'h'),
    ('lULParamValueListIndex', 'i'),
    ('sUnused', '52s'),
]


# Types of epoch (see head of file for description)
EPOCH_DISABLED = 0
EPOCH_STEPPED = 1
EPOCH_RAMPED = 2
EPOCH_RECTANGLE = 3
EPOCH_TRIANGLE = 4
EPOCH_COSINE = 5
EPOCH_UNUSED = 6     # Legacy issue: "was ABF_EPOCH_TYPE_RESISTANCE"
EPOCH_BIPHASIC = 7
epoch_types = {
    EPOCH_DISABLED: 'Disabled',
    EPOCH_STEPPED: 'Stepped waveform (square pulse)',
    EPOCH_RAMPED: 'Ramp waveform (fixed-angle in- or decrease)',
    EPOCH_RECTANGLE: 'Rectangular pulse train',
    EPOCH_TRIANGLE: 'Triangular waveform',
    EPOCH_COSINE: 'Cosine waveform',
    EPOCH_UNUSED: 'Unused',
    EPOCH_BIPHASIC: 'Biphasic pulse train',
}


# Fields in the epoch section (abf2)
# EpochInfoDescription = [
#       ('nEpochNum', 'h'),
#       ('nDigitalValue', 'h'),
#       ('nDigitalTrainValue', 'h'),
#       ('nAlternateDigitalValue', 'h'),
#       ('nAlternateDigitalTrainValue', 'h'),
#       ('bEpochCompression', 'b'),
#       ('sUnused', '21s'),
#   ]
ACMODE_VARIABLE_LENGTH_EVENTS = 1
ACMODE_FIXED_LENGTH_EVENTS = 2
ACMODE_GAP_FREE = 3
ACMODE_HIGH_SPEED_OSCILLOSCOPE = 4
ACMODE_EPISODIC_STIMULATION = 5
acquisition_modes = {
    # Variable length sweeps, triggered by some event
    ACMODE_VARIABLE_LENGTH_EVENTS: 'Variable-length events mode',
    # Fixed length sweeps, triggered by some event, may overlap
    ACMODE_FIXED_LENGTH_EVENTS: 'Event-driven fixed-length mode',
    # Continuous recording
    ACMODE_GAP_FREE: 'Gap free mode',
    # Fixed length non-overlapping, sweeps, triggered by some event
    ACMODE_HIGH_SPEED_OSCILLOSCOPE: 'High-speed oscilloscope mode',
    # Fixed length non-overlapping sweeps
    ACMODE_EPISODIC_STIMULATION: 'Episodic stimulation mode'
}


# DAC channel types
TYPE_UNKNOWN = 0
TYPE_VOLTAGE_CLAMP = 1
TYPE_CURRENT_CLAMP = 2
TYPE_CURRENT_CLAMP_ZERO = 4
type_modes = {
    0: TYPE_UNKNOWN,
    1: TYPE_VOLTAGE_CLAMP,
    2: TYPE_CURRENT_CLAMP,
    4: TYPE_CURRENT_CLAMP_ZERO,
}
type_mode_names = {
    0: 'Unknown',
    1: 'Voltage clamp',
    2: 'Current clamp',
    4: 'Current clamp zero',
}


# DAC waveform types
DAC_DISABLED = 0
DAC_EPOCHTABLEWAVEFORM = 1  # Epochs
DAC_DACFILEWAVEFORM = 2     # Stimulus file


# User list parameter to vary
'''
CONDITNUMPULSES 0
CONDITBASELINEDURATION 1
CONDITBASELINELEVEL 2
CONDITSTEPDURATION 3
CONDITSTEPLEVEL 4
CONDITPOSTTRAINDURATION 5
CONDITPOSTTRAINLEVEL 6
EPISODESTARTTOSTART 7
INACTIVEHOLDING 8
DIGITALINTEREPISODE 9
PNNUMPULSES 10
PARALLELVALUE(0-9) 11-20
EPOCHINITLEVEL(0-9) 21-30
EPOCHINITDURATION(0-9) 31-40
EPOCHTRAINPERIOD(0-9) 41-50
EPOCHTRAINPULSEWIDTH(0-9) 51-60
'''



#
# Non-standard exceptions raised by myokit.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Base classes
#
#
class MyokitError(Exception):
    """
    Base class for all exceptions specific to Myokit.

    Note that myokit classes and functions may raise any type of exception, for
    example a :class:``KeyError`` or a :class:`ValueError`. Only new classes of
    exception *defined* by Myokit will extend this base class.

    *Extends:* ``Exception``
    """
    def __init__(self, message):
        super().__init__(message)


class IntegrityError(MyokitError):
    """
    Raised if an integrity error is found in a model.

    The error message is stored in the property ``message``. An optional parser
    token may be obtained with :meth:`token()`.

    *Extends:* :class:`myokit.MyokitError`
    """
    def __init__(self, message, token=None):
        super().__init__(message)
        self._token = token

    def token(self):
        """
        Returns a parser token associated with this error, or ``None`` if no
        such token was given.
        """
        return self._token


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Inheriting classes
#
class CompilationError(MyokitError):
    """
    Raised if an auto-compiling class fails to compile. Catching one of these
    is usually a good excuses to email the developers ;-)

    *Extends:* :class:`myokit.MyokitError`
    """


class CyclicalDependencyError(IntegrityError):
    """
    Raised when an variables depend on each other in a cyclical manner.

    The first argument ``cycle`` must be a sequence containing the
    :class:`Variable` objects in the cycle.

    *Extends:* :class:`myokit.IntegrityError`
    """
    def __init__(self, cycle):
        # Set message
        msg = 'Cyclical reference found: (' + ' > '.join(
            [v.var().qname() for v in cycle]) + ').'
        # Set token: First item in cycle is model's defining lhs
        tok = cycle[0].var()._token
        # Raise
        super().__init__(msg, tok)


class DataBlockReadError(MyokitError):
    """
    Raised when an error is encountered while reading a
    :class:`myokit.DataBlock1d` or :class:`myokit.DataBlock2d`.

    *Extends:* :class:`myokit.MyokitError`.
    """


class DataLogReadError(MyokitError):
    """
    Raised when an error is encountered while reading a
    :class:`myokit.DataLog`.

    *Extends:* :class:`myokit.MyokitError`.
    """


class DuplicateName(MyokitError):
    """
    Raised when an attempt is made to add a component or variable with a name
    that is already in use within the relevant scope.

    *Extends:* :class:`myokit.MyokitError`.
    """


class DuplicateFunctionName(MyokitError):
    """
    Raised when an attempt is made to add a user function to a model when a
    function with the same name and number of arguments already exists.

    *Extends:* :class:`myokit.MyokitError`.
    """


class DuplicateFunctionArgument(MyokitError):
    """
    Raised when an attempt is made to define a user function with duplicate
    argument names.

    *Extends:* :class:`myokit.MyokitError`.
    """


class ExportError(MyokitError):
    """
    Raised when an export to another format fails.

    *Extends:* :class:`myokit.MyokitError`.
    """


class FindNanError(MyokitError):
    """
    Raised by some simulations when a search for the origins of a numerical
    error has failed.

    *Extends:* :class:`myokit.MyokitError`
    """


class GenerationError(MyokitError):
    """
    Raised by simulation engines and other auto-compiled classes if code
    generation fails.

    *Extends:* :class:`myokit.MyokitError`
    """


class IllegalAliasError(MyokitError):
    """
    Raised when an attempt is made to add an alias in an invalid manner.

    *Extends:* :class:`myokit.MyokitError`
    """


class IllegalReferenceError(IntegrityError):
    """
    Raised when a reference is found to a variable ``reference`` that isn't
    accessible from the owning variable ``owner``'s scope.

    *Extends:* :class:`myokit.IntegrityError`
    """
    def __init__(self, reference, owner):
        super().__init__(
            'Illegal reference: The referenced variable <' + reference.qname()
            + '> is outside the scope of <' + owner.qname() + '>.')


class IllegalReferenceInInitialValueError(IllegalReferenceError):
    """
    Raised when an illegal reference is made in an initial value.

    The only way this can occur is if the reference is to a nested variable.
    """
    def __init__(self, reference, owner):
        super(IntegrityError, self).__init__(
            'Illegal reference made in initial value: The referenced variable'
            ' <' + reference.qname() + '> is nested.')


class ImportError(MyokitError):
    """
    Raised when an import from another format fails.

    *Extends:* :class:`myokit.MyokitError`.
    """


class IncompatibleModelError(MyokitError):
    """
    Raised if a model is not compatible with some requirement.

    *Extends:* :class:`myokit.MyokitError`.
    """
    def __init__(self, model_name, message):
        msg = 'Incompatible model'
        if model_name:
            msg += ' <' + str(model_name) + '>'
        msg += ': ' + str(message)
        super().__init__(msg)


class IncompatibleUnitError(MyokitError):
    """
    Raised when a unit incompatibility is detected.

    *Extends:* :class:`myokit.MyokitError`.
    """
    def __init__(self, message, token=None):
        super().__init__(message)
        self._token = token

    def token(self):
        """
        Returns a parser token associated with this error, or ``None`` if no
        such token was given.
        """
        return self._token


class InvalidBindingError(IntegrityError):
    """
    Raised when an invalid binding is made.

    *Extends:* :class:`myokit.IntegrityError`
    """


class InvalidDataLogError(MyokitError):
    """
    Raised during validation of a :class:`myokit.DataLog` if a violation is
    found.

    *Extends:* :class:`myokit.MyokitError`.
    """


class InvalidFunction(MyokitError):
    """
    Raised when a function is declared with invalid arguments or an invalid
    expression.

    *Extends:* :class:`myokit.MyokitError`.
    """


class InvalidLabelError(IntegrityError):
    """
    Raised when an invalid label is set.

    *Extends:* :class:`myokit.IntegrityError`
    """


class InvalidNameError(MyokitError):
    """
    Raised when an attempt is made to add a component or variable with a name
    that violates the myokit naming rules.

    *Extends:* :class:`myokit.MyokitError`
    """


class InvalidMetaDataNameError(MyokitError):
    """
    Raised when an attempt is made to add a meta data property with a name
    that violates that the myokit naming rules for meta data properties.

    *Extends:* :class:`myokit.MyokitError`
    """


class MissingRhsError(IntegrityError):
    """
    Raised when a variable was declared without a defining right-hand side
    equation.

    The first argument ``var`` should be the invalid variable.

    *Extends:* :class:`myokit.IntegrityError`
    """
    def __init__(self, var):
        msg = 'No rhs set for <' + var.qname() + '>.'
        tok = var._token
        super().__init__(msg, tok)


class MissingTimeVariableError(IntegrityError):
    """
    Raised when no variable was bound to time.

    *Extends:* :class:`myokit.IntegrityError`
    """
    def __init__(self):
        msg = 'No variable bound to time. At least one of the model\'s' \
              ' variables must be bound to "time".'
        super().__init__(msg)


class NumericalError(MyokitError):
    """
    Raised when a numerical error occurs during the evaluation of a myokit
    :class:`Expression`.

    *Extends:* :class:`myokit.MyokitError`
    """


class ParseError(MyokitError):
    """
    Raised if an error is encountered during a parsing operation.

    A ParseError has five attributes:

    ``name``
        A short name describing the error
    ``line``
        The line the error occurred on (integer, first line is one)
    ``char``
        The character the error ocurred on (integer, first char is zero)
    ``desc``
        A more detailed description of the error.
    ``cause``
        Another exception that triggered this exception (or ``None``).

    *Extends:* :class:`myokit.MyokitError`
    """
    def __init__(self, name, line, char, desc, cause=None):
        self.name = str(name)
        self.line = int(line)
        self.char = int(char)
        self.value = self.name + ' on line ' + str(self.line)
        self.desc = str(desc)
        self.value += ': ' + self.desc
        self.cause = cause
        super().__init__(self.value)


class ProtocolEventError(MyokitError):
    """
    Raised when a :class:`ProtocolEvent` is created with invalid parameters.

    *Extends:* :class:`myokit.MyokitError`
    """


class ProtocolParseError(ParseError):
    """
    Raised when protocol parsing fails.

    *Extends:* :class:`ParseError`
    """


class SectionNotFoundError(MyokitError):
    """
    Raised if a section should be present in a file but is not.

    *Extends:* :class:`myokit.MyokitError`
    """


class SimulationError(MyokitError):
    """
    Raised when a numerical error occurred during a simulation. Contains a
    detailed error message.

    *Extends:* :class:`myokit.MyokitError`
    """
    # Only for numerical errors!


class SimulationCancelledError(MyokitError):
    """
    Raised when a user terminates a simulation.

    *Extends:* :class:`myokit.MyokitError`
    """
    def __init__(self, message='Operation cancelled by user.'):
        super().__init__(message)


class SimultaneousProtocolEventError(MyokitError):
    """
    Raised if two events in a protocol happen at the same time. Raised when
    creating a protocol or when running one.

    *Extends:* :class:`myokit.MyokitError`
    """


class UnresolvedReferenceError(IntegrityError):
    """
    Raised when a reference to a variable cannot be resolved.

    *Extends:* :class:`myokit.IntegrityError`
    """
    def __init__(self, reference, extra_message=None):
        super().__init__(
            'Unknown variable: <' + reference + '>.'
            + ((' ' + extra_message) if extra_message else '')
        )


class UnusedVariableError(IntegrityError):
    """
    Raised when an unused variable is found.

    The unused variable must be passed in as the first argument ``var``.

    *Extends:* :class:`myokit.IntegrityError`
    """
    def __init__(self, var):
        msg = 'Unused variable: <' + var.qname() + '>.'
        tok = var.lhs()._token
        super().__init__(msg, tok)


class VariableMappingError(MyokitError):
    """
    Raised when a method needs to map variables from one model onto another,
    but no valid mapping can be made.

    *Extends:* :class:`myokit.MyokitError`.
    """



#
# This module reads files in Axon Text File format.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import re

import numpy as np

from collections import OrderedDict

import myokit

# Format used for any text
_ENC = 'ascii'


class AtfFile:
    """
    Represents an Axon Text File (ATF) stored on disk.

    This method provides access to the data stored in the ATF as well as any
    meta data stored in the header.

    Access to the data is provided using a dict-like interface: to iterate over
    the file's keys use :meth:`iterkeys`, to select a value use
    ``atf_file['key']``. All iterators return the keys stored in the order they
    were listed in the ATF file.
    """
    def __init__(self, filename):
        # The path to the file and its basename
        self._path = os.path.abspath(filename)
        self._filename = os.path.basename(filename)

        # A version string
        self._version_str = None

        # A (multi-line) string containing meta-data found in this file
        self._meta = None

        # An ordered dict with key-value pairs. The first key is time.
        self._data = None

        # Read data
        self._read()

    def __getitem__(self, key):
        return self._data.__getitem__(key)

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def filename(self):
        """ Returns this ATF's filename. """
        return self._filename

    def info(self):
        """ Returns this ATF's header/meta data. """
        # Deprecated since 2023-07-12
        import warnings
        warnings.warn(
            'The method `info` is deprecated. Please use `meta_str` instead.')
        return self.meta_str()

    def items(self):
        """ Returns an iterator over all ``(key, value)`` pairs. """
        return iter(self._data.items())

    def keys(self):
        """ Returns an iterator over all keys in this ATF. """
        return iter(self._data.keys())

    def log(self):
        """
        Returns this file's time series data as a :class:`myokit.DataLog`.
        """
        log = myokit.DataLog()
        if len(self._data) > 0:
            log.set_time_key(next(iter(self._data.keys())))
        for k, v in self._data.items():
            log[k] = v
        return log

    def meta_str(self, verbose=True):
        """
        Returns this ATF's header data as an unstructured multi-line string.

        Note that the ``verbose`` argument doesn't do anything, but provides
        compatibility with similar methods in other files.
        """
        return self._meta

    def myokit_log(self):
        """ Deprecated alias of :meth:`log`. """
        # Deprecated since 2023-06-22
        import warnings
        warnings.warn(
            'The method `myokit_log` is deprecated. Please use `log` instead.')
        return self.log()

    def path(self):
        """ Returns the path to this ATF file. """
        return self._path

    def _read(self):
        """ Reads the data in this file. """
        with open(self._path, 'rb') as f:
            # Check version
            line = f.readline().decode(_ENC)
            line_index = 1
            if line[:3] != 'ATF':
                raise Exception('Unrecognised file type.')
            self._version_str = line[3:].strip()

            # Read number of header lines, number of fields
            line = f.readline().decode(_ENC)
            line_index += 1
            nh, nf = [int(x) for x in line.split()]

            # Read header data
            # If formatted as key-value pairs, format the meta data nicely.
            # Otherwise, just enter as is.
            key = []    # Keys
            val = []    # Values
            raw = []    # Fallback
            key_value_pairs = True
            for i in range(nh):
                line = f.readline().decode(_ENC).strip()
                line_index += 1
                if line[0] != '"' or line[-1] != '"':
                    raise Exception(
                        f'Invalid header on line f{line_index}: expecting '
                        f' lines wrapped in double quotes ("like this").')
                line = line[1:-1].strip()
                raw.append(line)
                if key_value_pairs:
                    try:
                        k, v = line.split('=')
                        key.append(k.strip())
                        val.append(v.strip())
                    except ValueError:
                        key_value_pairs = False
            if key_value_pairs:
                n = max([len(k) for k in key])
                meta = []
                val = iter(val)
                for k in key:
                    v = next(val)
                    meta.append(k + ' ' * (n - len(k)) + ' = ' + v)
                self._meta = '\n'.join(meta)
            else:
                self._meta = '\n'.join(raw)

            # Read time-series data
            self._data = OrderedDict()
            line = f.readline().decode(_ENC).strip()
            line_index += 1

            # Test if comma separated or space/tab separated
            delims = re.compile(r'["]{1}[^"]*["]{1}')
            delims = delims.split(line)

            # First and last delim must be empty (i.e. line starts and ends
            # with '"')
            if delims[0] != '' or delims[-1] != '':
                raise Exception('Unable to parse column headers.')
            delims = delims[1:-1]
            if len(delims) + 1 != nf:
                raise Exception(
                    f'Unable to parse column headers: Expected {nf} headers,'
                    f' found {len(delims) + 1}.')
            commas = ',' in delims[0]
            for delim in delims:
                if commas != (',' in delim):
                    raise Exception('Mixed delimiters are not supported.')

            # Read column headers
            keys = []
            try:
                a, b = 0, 0
                if line[a] != '"':  # pragma: no cover
                    # This should have triggered an earlier error
                    raise Exception('Unable to parse columns headers.')
                for i in range(nf):
                    b = line.index('"', a + 1)
                    keys.append(line[a + 1:b])
                    a = line.index('"', b + 1)
            except ValueError:
                pass
            if len(keys) != nf:     # pragma: no cover
                # This should have been picked up above
                raise Exception(
                    f'Unable to parse column headers: Expected {nf} headers,'
                    f' found {len(keys)}.')

            # Read data
            data = []
            for key in keys:
                col = []
                data.append(col)
                self._data[key] = col
            sep = ',' if commas else None
            for line in f:
                line_index += 1
                line = line.decode(_ENC).strip()
                vals = line.split(sep)
                if len(vals) != nf:
                    raise Exception(
                        f'Invalid data on line f{line_index}: expecting {nf}'
                        f' fields, found {len(vals)}.')
                vals = [float(x) for x in vals]
                for k, d in enumerate(vals):
                    data[k].append(d)

    def values(self):
        """ Returns an iterator over all values in this ATF. """
        return iter(self._data.values())

    def version(self):
        """ Returns a string representation of this file's version number. """
        return self._version_str


def load_atf(filename):
    """
    Reads an ATF file and returns its data as a :class:`myokit.DataLog`.
    """
    filename = os.path.expanduser(filename)
    return AtfFile(filename).log()


def save_atf(log, filename, fields=None):
    """
    Saves the :class:`myokit.DataLog` ``log`` to ``filename`` in ATF format.

    ATF requires that the times in the log be regularly spaced.

    The first column in an ATF file should always be time. Remaining fields
    will be written in a random order. To indicate an order or make a selection
    of fields, pass in a sequence ``fields`` containing the field names.
    """
    log.validate()
    import myokit

    # Check filename
    filename = os.path.expanduser(filename)

    # Delimiters
    # Dos-style EOL: Open file in 'wb' mode to stop windows writing \r\r\n
    eol = '\r\n'
    delim = '\t'

    # Create data and keys lists
    data = [iter(log.time())]
    time = log.time_key()
    keys = [time]

    # Check fields
    if fields:
        for field in fields:
            field = str(field)
            if field == time:
                continue
            keys.append(field)
            try:
                data.append(iter(log[field]))
            except KeyError:
                raise ValueError('Variable <' + field + '> not found in log.')
    else:
        for k, v in log.items():
            if k != time:
                keys.append(k)
                data.append(iter(v))

    for k in keys:
        if '"' in k:
            raise ValueError('Column names must not contain double quotes.')
        if '\r' in k or '\n' in k:
            raise ValueError(
                'Column names must not contain newlines or carriage returns.')

    # Check if time is equally spaced
    t = np.asarray(log.time())
    dt = t[1:] - t[:-1]
    dt_ref = dt[0]
    dt_err = dt_ref * 1e-6
    if np.any(np.abs(dt - dt_ref) > dt_err):
        raise ValueError('The time variable must be regularly spaced.')

    # Create header
    header = []
    header.append(('myokit-version', 'Myokit ' + myokit.version(raw=True)))
    header.append(('date-created', myokit.date()))
    header.append(('sampling-interval', dt_ref))

    # Get sizes
    nh = len(header)
    nf = len(keys)
    nd = log.length()

    # Write file
    with open(filename, 'wb') as f:
        # Write version number
        f.write(('ATF 1.0' + eol).encode(_ENC))

        # Write number of header lines, number of fields
        f.write(f'{nh}{delim}{nf}{eol}'.encode(_ENC))
        for k, v in header:
            f.write(f'"{k}={v}"{eol}'.encode(_ENC))

        # Write field names
        f.write((delim.join([f'"{k}"' for k in keys]) + eol).encode(_ENC))

        # Write data
        for i in range(nd):
            f.write((
                delim.join([myokit.float.str(next(d)) for d in data]) + eol
            ).encode(_ENC))




#!/usr/bin/env python3
#
# Tests the parser
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit
import myokit.pype

from myokit.tests import TemporaryDirectory


class PypeTest(unittest.TestCase):

    def test_process_errors(self):
        # Test error handling in the ``process`` method.

        # Process method takes a dict
        e = myokit.pype.TemplateEngine()
        self.assertRaisesRegex(
            ValueError, 'dict', e.process, 'file.txt', [])

        with myokit.tools.capture():

            # Test not-a-file
            self.assertRaises(IOError, e.process, 'file.txt', {})

            # Test simple error
            self.e("""<?print(1/0) ?>""", {}, 'ZeroDivisionError')

            # Test closing without opening
            self.e("""Hello ?>""", {}, 'without opening tag')

            # Opening without closing is allowed
            self.e("""<?print('hi')""", {})

            # Nested opening
            self.e(
                """<?print('hi')<?print('hello')?>""",
                {},
                'Nested opening tag',
            )

            # Too much inside <?=?>
            self.e("""<?=if 1 > 2: print('hi')?>""", {}, 'contain a single')
            self.e("""<?=print(1); print('hi')?>""", {}, 'contain a single')

            # Triple quote should be allowed
            self.e('''Hello"""string"""yes''', {})

            # OSError from inside pype
            self.e("""Hello<?open('file.txt', 'r')?>yes""", {}, 'No such file')

    def e(self, template, args, expected_error=None):
        """
        Runs a template, if an error is expected it checks if it's the right
        one.
        """
        with TemporaryDirectory() as d:
            path = d.path('template')
            with open(path, 'w') as f:
                f.write(template)
            e = myokit.pype.TemplateEngine()
            if expected_error is None:
                e.process(path, args)
            else:
                try:
                    e.process(path, args)
                except myokit.pype.PypeError:
                    # Check expected message in error details
                    self.assertIn(expected_error, e.error_details())
                    return
                raise RuntimeError('PypeError not raised.')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests the methods and classes in myokit.floats
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import sys
import unittest

import myokit


class FloatTest(unittest.TestCase):
    """Tests the methods in ``myokit.float``."""

    def test_close(self):
        # Test that close allows bigger errors than eq
        x = 49
        y = x * (1 + 1e-11)
        self.assertNotEqual(x, y)
        self.assertFalse(myokit.float.eq(x, y))
        self.assertTrue(myokit.float.close(x, y))

        # And that close thinks everything small is equal
        x = 1e-16
        y = 1e-12
        self.assertNotEqual(x, y)
        self.assertFalse(myokit.float.eq(x, y))
        self.assertTrue(myokit.float.close(x, y))

    def test_cround(self):
        # Test rounding based on closeness
        x = 49
        y = x * (1 + 1e-11)
        self.assertNotEqual(x, y)
        self.assertEqual(x, myokit.float.cround(y))
        self.assertIsInstance(myokit.float.cround(y), int)
        self.assertNotEqual(x, myokit.float.cround(49.001))

    def test_eq_and_geq(self):
        # Test the floating point comparison methods

        # Check that test is going to work
        x = 49
        y = 1 / (1 / x)
        self.assertNotEqual(x, y)

        # Test if eq allows 1 floating point error
        self.assertTrue(myokit.float.eq(x, y))
        self.assertTrue(myokit.float.geq(x, y))
        self.assertTrue(myokit.float.geq(y, x))
        x += x * sys.float_info.epsilon
        self.assertTrue(myokit.float.eq(x, y))
        self.assertTrue(myokit.float.geq(x, y))
        self.assertTrue(myokit.float.geq(y, x))
        x += x * sys.float_info.epsilon
        self.assertFalse(myokit.float.eq(x, y))
        self.assertTrue(myokit.float.geq(x, y))
        self.assertFalse(myokit.float.geq(y, x))

    def test_round(self):
        # Test rounding, ignoring single rounding errors

        # Check that test is going to work
        x = 49
        y = 1 / (1 / x)
        self.assertNotEqual(x, y)
        x += 2 * x * sys.float_info.epsilon
        self.assertFalse(myokit.float.eq(x, y))
        self.assertTrue(myokit.float.geq(x, y))
        self.assertFalse(myokit.float.geq(y, x))

        # Test rounding
        self.assertNotEqual(49, y)
        self.assertEqual(49, myokit.float.round(y))
        self.assertNotEqual(49, myokit.float.round(x))
        self.assertEqual(0.5, myokit.float.round(0.5))
        self.assertIsInstance(myokit.float.round(y), int)

        # Try with negative numbers
        self.assertNotEqual(-49, -y)
        self.assertEqual(-49, myokit.float.round(-y))
        self.assertNotEqual(-49, myokit.float.round(-x))
        self.assertEqual(-0.5, myokit.float.round(-0.5))

    def test_str(self):
        # Test float to string conversion.

        # String should be passed through
        self.assertEqual(myokit.float.str('123'), '123')

        # Simple numbers
        self.assertEqual(myokit.float.str(0), '0')
        self.assertEqual(myokit.float.str(0.0000), '0.0')
        self.assertEqual(myokit.float.str(1.234), '1.234')
        self.assertEqual(
            myokit.float.str(0.12432656245e12), ' 1.24326562450000000e+11')
        self.assertEqual(myokit.float.str(-0), '0')
        self.assertEqual(myokit.float.str(-0.0000), '-0.0')
        self.assertEqual(myokit.float.str(-1.234), '-1.234')
        self.assertEqual(
            myokit.float.str(-0.12432656245e12), '-1.24326562450000000e+11')

        # Strings are not converted
        x = '1.234'
        self.assertEqual(x, myokit.float.str(x))

        # Myokit Numbers are converted
        x = myokit.Number(1.23)
        self.assertEqual(myokit.float.str(x), '1.23')

        # Single and double precision
        self.assertEqual(
            myokit.float.str(-1.234, precision=myokit.SINGLE_PRECISION),
            '-1.234')
        self.assertEqual(
            myokit.float.str(
                -0.124326562458734682153498731245756e12,
                precision=myokit.SINGLE_PRECISION),
            '-1.243265625e+11')
        self.assertEqual(
            myokit.float.str(-1.234, precision=myokit.DOUBLE_PRECISION),
            '-1.234')
        self.assertEqual(
            myokit.float.str(
                -0.124326562458734682153498731245756e12,
                precision=myokit.DOUBLE_PRECISION),
            '-1.24326562458734680e+11')

        # Full precision override
        self.assertEqual(
            myokit.float.str(1.23, True), ' 1.22999999999999998e+00')
        self.assertEqual(
            myokit.float.str(1.23, True, myokit.DOUBLE_PRECISION),
            ' 1.22999999999999998e+00')
        self.assertEqual(
            myokit.float.str(1.23, True, myokit.SINGLE_PRECISION),
            ' 1.230000000e+00')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests the Jacobian tracer.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import myokit

from myokit.tests import DIR_DATA


class JacobianTracerTest(unittest.TestCase):
    """
    Tests the JacobianTracer.
    """
    def test_simple(self):
        # Load model
        m = os.path.join(DIR_DATA, 'lr-1991.mmt')
        m, p, x = myokit.load(m)
        v = m.binding('diffusion_current')
        if v is not None:
            v.set_binding(None)

        # Run a simulation, save all states & bound values
        s = myokit.Simulation(m, p)
        s.pre(10)
        s.reset()
        d = s.run(20, log=myokit.LOG_STATE + myokit.LOG_BOUND)

        # Calculate jacobians from log
        g = myokit.JacobianTracer(m)
        b = g.jacobians(d)

        # Calculate the dominant eigenvalues
        g.dominant_eigenvalues(log=d)
        g.dominant_eigenvalues(block=b)
        self.assertRaises(ValueError, g.dominant_eigenvalues)

        # Calculate the largest eigenvalues
        g.largest_eigenvalues(log=d)
        g.largest_eigenvalues(block=b)
        self.assertRaises(ValueError, g.largest_eigenvalues)

        # Log missing a state
        d2 = d.clone()
        del d2['membrane.V']
        self.assertRaisesRegex(ValueError, 'all state', g.jacobians, d2)

        # Log entries differ in length
        d2['membrane.V'] = d['membrane.V'][:-1]
        self.assertRaisesRegex(ValueError, 'same length', g.jacobians, d2)

        # Log missing time
        d2['membrane.V'] = d['engine.time']
        del d2['engine.time']
        self.assertRaisesRegex(ValueError, 'bound', g.jacobians, d2)


if __name__ == '__main__':
    unittest.main()



#
# Provides support for working with data in formats used by Axon Instruments.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from ._abf import AbfFile, Sweep, Channel  # noqa
from ._atf import AtfFile, load_atf, save_atf  # noqa
from ._importer import AbfImporter


# Importers
_importers = {
    'abf': AbfImporter,
}


def importers():
    """
    Returns a dict of all importers available in this module.
    """
    return dict(_importers)
# Exporters
# Expression writers



#
# Stripped version of myokit.DataLog
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import array
from collections import OrderedDict


class DataLog(OrderedDict):
    """
    Extends an ordered ``dict`` type with a number of methods specifically for
    working with logged simulation data.
    """
    def get(self, name, *cell):
        """
        Convenience method for access to multi-cell data. For example, in a 0d
        (single cell) simulation, a log could look like this::

            d = myokit.DataLog()
            d['engine.time'] = numpy.range(0, 10)
            d['membrane.V'] = 10 * d['engine.time']

        In this case, the ``get`` method just returns whatever is stored at the
        given name::

            d.get('membrane.V') == d['membrane.V']

        In a 1d cable situation, a log can be built like this::

            d = myokit.DataLog()
            d['engine.time'] = numpy.range(0, 10)
            d['0.membrane.V'] = 10 * d['engine.time']       # First cell
            d['1.membrane.V'] = 20 * d['engine.time']       # Second cell

        In this case, the ``get`` method can be used to access the different
        cells::

            d.get('membrane.V', 0) == d['0.membrane.V']

        Similarly, in a 2d tissue log ``d.get('membrane.V', 1, 2)`` will return
        the entry stored at ``d['1.2.membrane.V']``.
        """
        key = [str(x) for x in cell]
        key.append(str(name))
        key = '.'.join(key)
        return self[key]

    @staticmethod
    def load_csv(filename):
        """
        Loads a CSV file from disk and returns it as a :class:`DataLog`.

        The CSV file must start with a header line indicating the variable
        names, separated by commas. Each subsequent row should contain the
        values at a single point in time for all logged variables.
        """
        # Check filename
        filename = os.path.expanduser(filename)

        # Error method
        def e(line, char, msg):
            raise Exception(
                'Syntax error on line ' + str(line) + ', character '
                + str(1 + char) + ': ' + msg)

        # Delimiters
        quote = '"'
        delim = ','
        with open(filename, 'r', newline=None) as f:
            # Read header
            keys = []
            try:
                line = f.readline()
            except EOFError:
                e(0, 0, 'Empty file, expecting header.')
            # Trim end of line
            if len(line) > 1 and line[-2:] == '\r\n':
                eol = 2
                line = line[:-2]
            else:
                eol = 1
                line = line[:-1]
            # Trim ; at end of line if given
            if line[-1] == ';':
                eol += 1
                line = line[:-1]
            # Get enumerated iterator over characters
            line = enumerate(line)
            try:
                i, c = next(line)
            except StopIteration:
                e(1, i, 'Empty line, expecting header.')
            run1 = True
            while run1:
                text = []
                if c == quote:
                    # Read quoted field + delimiter or eol
                    run2 = True
                    while run2:
                        try:
                            i, c = next(line)
                        except StopIteration:
                            e(
                                1, i,
                                'Unexpected end-of-line inside quoted string.')
                        if c == quote:
                            try:
                                i, c = next(line)
                                if c == quote:
                                    text.append(quote)
                                elif c == delim:
                                    run2 = False
                                else:
                                    e(
                                        1, i, 'Expecting double quote,'
                                        ' delimiter or end-of-line. Found "'
                                        + c + '".')
                            except StopIteration:
                                run1 = run2 = False
                        else:
                            text.append(c)
                else:
                    # Read unquoted field + delimiter or eol
                    while run1 and c != delim:
                        try:
                            text.append(c)
                            i, c = next(line)
                        except StopIteration:
                            run1 = False
                # Append new field to list
                keys.append(''.join(text))
                # Read next character
                try:
                    i, c = next(line)
                except StopIteration:
                    run1 = False
            if c == delim:
                e(1, i, 'Empty field in header.')
            # Create data structure
            m = len(keys)
            log = DataLog()
            lists = []
            for key in keys:
                x = array.array('d')
                lists.append(x)
                log[key] = x
            # Read remaining data
            try:
                n = 0
                while True:
                    row = f.readline()
                    if row.strip() == '':
                        break
                    row = row[:-eol]
                    row = row.split(delim)
                    n += 1
                    if len(row) != m:
                        raise Exception(
                            'Wrong number of columns found in row ' + str(n)
                            + '. Expecting ' + str(m) + ', found '
                            + str(len(row)) + '.')
                    for k, v in enumerate(row):
                        lists[k].append(float(v))
            except StopIteration:
                pass
            # Return log
            return log

    def npview(self):
        """
        Returns an ordered dict of numpy arrays pointing to the data stored in
        this log.
        """
        import numpy as np
        out = DataLog()
        for k, d in self.items():
            out[k] = np.array(d, copy=False)
        return out

    def save_csv(self, filename, pad=None):
        """
        Writes this ``DataLog`` to a CSV file, following the syntax
        outlined in RFC 4180 and with a header indicating the field names.

        The resulting file will consist of:

          - A header line containing the names of all logged variables,
            separated by commas.
          - Each following line will be a comma separated list of values in the
            same order as the header line. A line is added for each time point
            logged.
        """
        # Check filename
        filename = os.path.expanduser(filename)
        # EOL: CSV files use DOS line endings '\r\n'. In windows, if you try
        # to write a '\n' in mode 'w' it automatically writes '\r\n', so '\r\n'
        # gets converted to '\r\r\n'. To circumvent this, open the file in mode
        # 'wb'.
        # Save
        eol = '\r\n'
        delim = ','
        quote = '"'
        escape = '""'
        with open(filename, 'wb') as f:
            # Convert dict structure to ordered sequences
            keys = []
            data = []
            n = []
            for key, dat in sorted(self.items()):
                keys.append(key)
                data.append(iter(dat))
                n.append(len(dat))
            n = set(n)
            if len(n) > 1:
                # Padding needed, check if provided
                if pad is None:
                    raise Exception(
                        'Data passed to save_csv contains lists of unequal'
                        ' length (' + str(n) + '). Please ensure all lists'
                        ' have equal length or set a padding value.')
            else:
                # Padding not needed
                pad = None
            n = max(n)
            # Write header
            line = []
            for key in keys:
                # Escape quotes within strings
                line.append(quote + key.replace(quote, escape) + quote)
            f.write(delim.join(line))
            f.write(eol)
            # Write data
            if pad is None:
                for i in range(0, n):
                    line = []
                    for d in data:
                        line.append(myokit.float.str(next(d)))
                    f.write(delim.join(line) + eol)
            else:
                for i in range(0, n):
                    line = []
                    for d in data:
                        try:
                            line.append(myokit.float.str(next(d)))
                        except StopIteration:
                            line.append('0')
                    f.write(delim.join(line) + eol)



#!/usr/bin/env python3
#
# Tests the protocol output for step protocols where floating point accuracy
# matters.
#
# In particular, for a protocol like:
#
# level     start       duration
# -80       0           3.3333
# -70       3.3333      3.3331
# -60       6.6664      3.3336
#
# We can run into problems with doubles, as:
#
#   >>> 3.3333 + 3.3331
#   6.666399999999999
#   >>> 3.3333 + 3.3331 < 6.6664
#   True
#
# This can cause the 2nd even to end just before the 3d starts, leading to a
# quick jump to 0 in between -70 and -60.
#
# To avoid this, we should use slightly more careful ways of comparing floating
# point numbers if all protocol handling code.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit

from myokit.tests.ansic_event_based_pacing import AnsicEventBasedPacing


class ProtocolFloatingPointTest(unittest.TestCase):

    def test_is_sequence_and_length_floats_1(self):

        # Tests how float issues are handled in characteristic_time,
        # is_sequence, and is_unbroken_sequence

        # Example 1: Sum too high

        # Good arithmetic: 1.2345 + 2.3454 = 3.5799
        # With doubles   :                   3.5799000000000003
        # So that        : 1.2345 + 2.3454 > 3.5799

        # Absolute error :                   4.440892098500626e-16
        # Float epsilon  : sys.float_info    2.220446049250313e-16
        # Abs error / eps:                   2.0 = 2**1
        # 2-log of sum   :                   [1-2]
        # Relative error : abs error / sum   1.2405073042544837e-16

        p = myokit.Protocol()
        p.schedule(-80, 0, 1.2345)
        p.schedule(-70, 1.2345, 2.3454)
        p.schedule(-60, 3.5799, 1.4201)

        self.assertEqual(p.characteristic_time(), 5)
        self.assertTrue(p.is_sequence())
        self.assertTrue(p.is_unbroken_sequence())

    def test_is_sequence_and_length_floats_2(self):

        # Tests how float issues are handled in characteristic_time,
        # is_sequence, and is_unbroken_sequence

        # Example 2: Sum too low

        # Good arithmetic: 3.3333 + 3.3331 = 6.6664
        # With doubles   :                   6.666399999999999
        # So that        : 3.3333 + 3.3331 < 6.6664

        # Absolute error :                   -8.881784197001252e-16
        # Float epsilon  : sys.float_info     2.220446049250313e-16
        # Abs error / eps:                   -4.0 = 2**2
        # 2-log of sum   :                    [2-3]
        # Relative error : abs error / sum   -1.3323209223870833e-16

        p = myokit.Protocol()
        p.schedule(-80, 0, 3.3333)
        p.schedule(-70, 3.3333, 3.3331)
        p.schedule(-60, 6.6664, 3.3336)

        self.assertEqual(p.characteristic_time(), 10)
        self.assertTrue(p.is_sequence())
        self.assertTrue(p.is_unbroken_sequence())

    def test_is_sequence_and_length_floats_3(self):

        # Tests how float issues are handled in characteristic_time,
        # is_sequence, and is_unbroken_sequence

        # Example 3: The size of the numbers matters
        #
        # Good arithmetic: 1340.6 + 22159.6 = 23500.2
        # With doubles   :                    23500.199999999997
        # So that        : 1340.6 + 22159.6 < 23500.2

        # Error          :                   -3.637978807091713e-12
        # Float epsilon  : sys.float_info     2.220446049250313e-16
        # Abs error / eps:                   -16384.0 = 2**14
        # 2-log of sum   :                    [14-15]
        # Relative error : abs error / sum   -1.5480629131206173e-16

        p = myokit.Protocol()
        p.schedule(-80, 0, 22159.6)
        p.schedule(-70, 22159.6, 1340.6)
        p.schedule(-60, 23500.2, 6499.8)

        self.assertEqual(p.characteristic_time(), 30000)
        self.assertTrue(p.is_sequence())
        self.assertTrue(p.is_unbroken_sequence())


class PacingSystemFloatingPointTest(unittest.TestCase):
    """
    Test float behavior in the Python PacingSystem implementation, using the
    log_for_interval method.
    """

    def test_python_pacing_floats_1(self):

        # Test fixes for event start/end touching
        p = myokit.Protocol()
        p.schedule(-80, 0, 1.2345)
        p.schedule(-70, 1.2345, 2.3454)
        p.schedule(-60, 3.5799, 1.4201)

        # Call log_for_interval, which uses the PacingSystem
        d = p.log_for_interval(0, 10)

        # Check if the correct values are recorded
        self.assertEqual(d['time'], [0, 1.2345, 3.5799, 5, 10])
        self.assertEqual(d['pace'], [-80, -70, -60, 0, 0])

        # Test fixes for start/end of interval touching events
        p = myokit.Protocol()
        p.schedule(-80, 0, 1.2345)
        p.schedule(-70, 1.2345, 2.3454)

        # Interval end
        d = p.log_for_interval(0, 3.5799)
        self.assertEqual(d['time'], [0, 1.2345, 3.5799])
        self.assertEqual(d['pace'], [-80, -70, 0])

        # Interval start
        d = p.log_for_interval(3.5799, 10)
        self.assertEqual(d['time'], [3.5799, 10])
        self.assertEqual(d['pace'], [0, 0])

    def test_python_pacing_floats_2(self):

        p = myokit.Protocol()
        p.schedule(-80, 0, 3.3333)
        p.schedule(-70, 3.3333, 3.3331)
        p.schedule(-60, 6.6664, 3.3336)

        # Call log_for_interval, which uses the PacingSystem
        d = p.log_for_interval(0, 20)

        # Check if the correct values are recorded
        self.assertEqual(d['time'], [0, 3.3333, 6.6664, 10, 20])
        self.assertEqual(d['pace'], [-80, -70, -60, 0, 0])

    def test_python_pacing_floats_3(self):

        p = myokit.Protocol()
        p.schedule(-80, 0, 22159.6)
        p.schedule(-70, 22159.6, 1340.6)
        p.schedule(-60, 23500.2, 6499.8)

        # Call log_for_interval, which uses the PacingSystem
        d = p.log_for_interval(0, 40000)

        # Check if the correct values are recorded
        self.assertEqual(d['time'], [0, 22159.6, 23500.2, 30000, 40000])
        self.assertEqual(d['pace'], [-80, -70, -60, 0, 0])


class AnsicPacingSystemFloatingPointTest(unittest.TestCase):
    """
    Test floating point behavior of C-pacing, using the Python wrapper and the
    method log_for_interval.
    """

    def test_C_pacing_floats_1(self):

        p = myokit.Protocol()
        p.schedule(-80, 0, 1.2345)
        p.schedule(-70, 1.2345, 2.3454)
        p.schedule(-60, 3.5799, 1.4201)

        # Call log_for_interval, which uses the PacingSystem
        d = AnsicEventBasedPacing.log_for_interval(p, 0, 10)

        # Check if the correct values are recorded
        self.assertEqual(d['time'], [0, 1.2345, 3.5799, 5, 10])
        self.assertEqual(d['pace'], [-80, -70, -60, 0, 0])

        # Test fixes for start/end of interval touching events
        p = myokit.Protocol()
        p.schedule(-80, 0, 1.2345)
        p.schedule(-70, 1.2345, 2.3454)

        # Interval end
        d = p.log_for_interval(0, 3.5799)
        self.assertEqual(d['time'], [0, 1.2345, 3.5799])
        self.assertEqual(d['pace'], [-80, -70, 0])

        # Interval start
        d = p.log_for_interval(3.5799, 10)
        self.assertEqual(d['time'], [3.5799, 10])
        self.assertEqual(d['pace'], [0, 0])

    def test_C_pacing_floats_2(self):

        p = myokit.Protocol()
        p.schedule(-80, 0, 3.3333)
        p.schedule(-70, 3.3333, 3.3331)
        p.schedule(-60, 6.6664, 3.3336)

        # Call log_for_interval, which uses the PacingSystem
        d = AnsicEventBasedPacing.log_for_interval(p, 0, 20)

        # Check if the correct values are recorded
        self.assertEqual(d['time'], [0, 3.3333, 6.6664, 10, 20])
        self.assertEqual(d['pace'], [-80, -70, -60, 0, 0])

    def test_C_pacing_floats_3(self):

        p = myokit.Protocol()
        p.schedule(-80, 0, 22159.6)
        p.schedule(-70, 22159.6, 1340.6)
        p.schedule(-60, 23500.2, 6499.8)

        # Call log_for_interval, which uses the PacingSystem
        d = AnsicEventBasedPacing.log_for_interval(p, 0, 40000)

        # Check if the correct values are recorded
        self.assertEqual(d['time'], [0, 22159.6, 23500.2, 30000, 40000])
        self.assertEqual(d['pace'], [-80, -70, -60, 0, 0])


class CVodeSimulationFloatingPointTest(unittest.TestCase):
    """
    Test floating point behavior of C-pacing, using the myokit Simulation class
    with a model without protocol (cvode-free mode).
    """

    def test_cvode_floating_point_protocol_1(self):

        # Tests the protocol handling in a CVODE simulation, which uses the
        # pacing.h file shared by all C/C++ simulation code.

        # Create model without states
        m = myokit.Model()
        c = m.add_component('c')
        t = c.add_variable('t')
        t.set_rhs(0)
        t.set_binding('time')
        v = c.add_variable('v')
        v.set_rhs('0')
        v.set_binding('pace')

        # Create tricky protocol
        p = myokit.Protocol()
        p.schedule(-80, 0, 1.2345)
        p.schedule(-70, 1.2345, 2.3454)
        p.schedule(-60, 3.5799, 1.4201)

        # Run & test
        s = myokit.Simulation(m, p)
        d = s.run(p.characteristic_time())
        self.assertEqual(list(d['c.v']), [-80, -70, -60, 0])

        # Test starting/stopping at difficult points
        p = myokit.Protocol()
        p.schedule(-80, 0, 1.2345)
        p.schedule(-70, 1.2345, 2.3454)

        # Interval end
        s = myokit.Simulation(m, p)
        d = s.run(3.5799)
        self.assertEqual(list(d['c.t']), [0, 1.2345, 3.5799])
        self.assertEqual(list(d['c.v']), [-80, -70, 0])

        # Interval start
        d = s.run(6.4201)
        self.assertEqual(list(d['c.t']), [3.5799, 10])
        self.assertEqual(list(d['c.v']), [0, 0])

    def test_cvode_floating_point_protocol_2(self):

        # Tests the protocol handling in a CVODE simulation, which uses the
        # pacing.h file shared by all C/C++ simulation code.

        m = myokit.Model()
        c = m.add_component('c')
        t = c.add_variable('t')
        t.set_rhs(0)
        t.set_binding('time')
        v = c.add_variable('v')
        v.set_rhs('0')
        v.set_binding('pace')

        p = myokit.Protocol()
        p.schedule(-80, 0, 3.3333)
        p.schedule(-70, 3.3333, 3.3331)
        p.schedule(-60, 6.6664, 3.3336)

        s = myokit.Simulation(m, p)
        d = s.run(10)

        self.assertEqual(list(d['c.t']), [0, 3.3333, 6.6664, 10])
        self.assertEqual(list(d['c.v']), [-80, -70, -60, 0])

    def test_cvode_floating_point_protocol_3(self):

        # Tests the protocol handling in a CVODE simulation, which uses the
        # pacing.h file shared by all C/C++ simulation code.

        m = myokit.Model()
        c = m.add_component('c')
        t = c.add_variable('t')
        t.set_rhs(0)
        t.set_binding('time')
        v = c.add_variable('v')
        v.set_rhs('0')
        v.set_binding('pace')

        p = myokit.Protocol()
        p.schedule(-80, 0, 22159.6)
        p.schedule(-70, 22159.6, 1340.6)
        p.schedule(-60, 23500.2, 6499.8)

        # Test without state variables, dynamic logging
        s = myokit.Simulation(m, p)
        d = s.run(40000)
        self.assertEqual(list(d['c.t']), [0, 22159.6, 23500.2, 30000, 40000])
        self.assertEqual(list(d['c.v']), [-80, -70, -60, 0, 0])

        # Test without state variables, fixed log times
        s.reset()
        d = s.run(40001, log_times=d['c.t'])
        self.assertEqual(list(d['c.t']), [0, 22159.6, 23500.2, 30000, 40000])
        self.assertEqual(list(d['c.v']), [-80, -70, -60, 0, 0])

        x = c.add_variable('x')
        x.set_rhs('cos(t)')
        x.promote(0)

        # Test with state variables, fixed log times
        s = myokit.Simulation(m, p)
        d = s.run(40001, log_times=d['c.t'])
        self.assertEqual(list(d['c.t']), [0, 22159.6, 23500.2, 30000, 40000])
        self.assertEqual(list(d['c.v']), [-80, -70, -60, 0, 0])

    def test_ending_near_protocol_end(self):
        # Tests ending a CVODE simulation at a point almost equal to (i.e.
        # indistinguishable from) an event point

        m = myokit.load_model('example')
        m.binding('pace').set_binding(None)
        v = m.label('membrane_potential')
        v.set_rhs(0)
        v.demote()
        v.set_binding('pace')

        p = myokit.Protocol()
        p.schedule(-80, 0, 15400)

        # This will cause an error unles handled explicitly
        s = myokit.Simulation(m, p)
        s.run(15400.000000000002)


if __name__ == '__main__':
    unittest.main()
